#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"

import os
import sys
import tarfile
import shutil
import csv

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import HDIConfig
import Utility
from CommonLogger import log_debug, log_info, log_warning, log_error

sys.path.append(os.path.join(BASEDIR, "bin"))
import backup_sync_common

# Constants
FSCONF = "filesystems.xml"
HISTCONF = "history.csv"
CIFSCONF = "CIFS_%s.xml"
NFSCONF = "NFS_%s.xml"
MIGTASKCONF = "migtasks.xml"
LFTCONF = "arclftctl.xml"
RESIDENTCONF = "arcresidentctl.xml"
NETWORKCONF = "network.xml"
HOSTCONF = "hostname.txt"
UUIDCONF = "system_uuid.conf"

class BaseError(Exception):
    def __init__(self, reason, code=None):
        self.reason = reason.strip()
        self.code = code

    def __str__(self):
        if self.code is None:
            return "%s" % self.reason
        else:
            return "%s (code=%s)" % (self.reason, self.code)

class ParameterError(BaseError):
    pass

class ConfigurationError(BaseError):
    pass

class HDIConfigArchive(object):
    def __init__(self):
        self.fslist = []
        self.history_count = {}
        self.synced_fslist = {}
        self.cifslist = []
        self.nfslist = []
        self.tasklist = []
        self.residentreport = None
        self.iplist = []
        self.hostname = None
        self.uuid = None
        self.work_dir = None
        self.archive_path = None

    def prepare_save(self, work_dir):
        try:
            if os.path.exists(work_dir):
                shutil.rmtree(work_dir)
            os.mkdir(work_dir)
        except EnvironmentError, e:
            log_error("Failed to clean up a working directory. error=%s, directory=%s" % (str(e), work_dir))
            raise
        self.work_dir = work_dir

    def save_file_systems(self, xml):
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        to_file = os.path.join(self.work_dir, FSCONF)
        self.save_data_to_file(xml, to_file)

    def save_history_count(self, history_count):
        """ Save history count for each file system with CSV format.
        history_count is a dictionary where key is file system name and value
        is history count on the file system.
        """
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        to_file = os.path.join(self.work_dir, HISTCONF)
        try:
            with open(to_file, 'w') as f:
                writer = csv.writer(f, lineterminator='\n')
                for fsname in history_count.keys():
                    writer.writerow([fsname, history_count[fsname]])
        except EnvironmentError, e:
            log_error("Failed to save history count information to a file. file=%s, error=%s" % (to_file, str(e)))
            raise

    def save_synced_file_systems(self, synced_dir_path):
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        if not os.path.exists(synced_dir_path):
            # There is no synced directory. Do nothing
            log_info("Synced directory does not exist. path=%s" % synced_dir_path)
            return
        dir_name = os.path.basename(synced_dir_path)
        to_dir_path = os.path.join(self.work_dir, dir_name)
        log_info("Copying %s to %s ..." % (synced_dir_path, to_dir_path))
        try:
            shutil.copytree(synced_dir_path, to_dir_path)
        except EnvironmentError, e:
            log_error("Failed to copy synced directory to working directory. error=%s, from=%s, to=%s" % (str(e), synced_dir_path, to_dir_path))
            raise

    def save_cifs_shares(self, from_file, fsname):
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        to_file = os.path.join(self.work_dir, CIFSCONF % fsname)
        self.rename_file(from_file, to_file)

    def save_nfs_shares(self, from_file, fsname):
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        to_file = os.path.join(self.work_dir, NFSCONF % fsname)
        self.rename_file(from_file, to_file)

    def save_mig_tasks(self, xml):
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        to_file = os.path.join(self.work_dir, MIGTASKCONF)
        self.save_data_to_file(xml, to_file)

    def save_lft_setting(self, xml):
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        to_file = os.path.join(self.work_dir, LFTCONF)
        self.save_data_to_file(xml, to_file)

    def save_cache_resident_reporting(self, xml):
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        to_file = os.path.join(self.work_dir, RESIDENTCONF)
        self.save_data_to_file(xml, to_file)

    def save_network(self, xml):
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        to_file = os.path.join(self.work_dir, NETWORKCONF)
        self.save_data_to_file(xml, to_file)

    def save_hostname(self, hostname):
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        to_file = os.path.join(self.work_dir, HOSTCONF)
        self.save_data_to_file(hostname, to_file)

    def save_system_uuid(self, uuid):
        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        to_file = os.path.join(self.work_dir, UUIDCONF)
        self.save_data_to_file(uuid, to_file)

    def build(self, filename=None):
        ''' Build a tar archive with the saved files, and returns the name of the archive file. '''

        if self.work_dir is None:
            log_error("Working directory has not been prepared yet.")
            raise ParameterError("Preparation of a working directory is required before saving any information.")
        if filename is None:
            # Use random string as a file name if not specified.
            filename = Utility.random_string(16) + ".tar.gz"

        log_info("Building an HDI configuration archive file ...")

        # Get list of files to be added to the archive
        try:
            items = os.listdir(self.work_dir)
        except EnvironmentError, e:
            # self.work_dir must exist because it is created before self.work_dir is set.
            log_error("An error occurred while getting list of files under the working directory. error=%s, work_dir=%s" % (str(e), self.work_dir))
            raise

        # Archive the files
        archive_path = os.path.join(self.work_dir, filename)
        tf = None
        try:
            tf = tarfile.open(archive_path, 'w:gz')
            for item in items:
                item_path = os.path.join(self.work_dir, item)
                tf.add(item_path, arcname=item)
        except tarfile.TarError, e:
            log_error("An error occurred while adding a file to an archive file. error=%s, file=%s, archive=%s" % (str(e), item_path, archive_path))
            raise
        finally:
            if tf is not None:
                tf.close()

        log_info("HDI configuration archive file was built as %s." % archive_path)

        return archive_path

    def extract_cifs_xml_file(self, fsname, to_dir):
        cifsfile = CIFSCONF % fsname
        return self.extract_file(cifsfile, to_dir)

    def extract_nfs_xml_file(self, fsname, to_dir):
        nfsfile = NFSCONF % fsname
        return self.extract_file(nfsfile, to_dir)

    def extract_file(self, filename, to_dir):
        log_info("Extracting %s from %s and put on %s ..." % (filename, self.archive_path, to_dir))
        if self.archive_path is None:
            log_warning("No file can be extracted because no archive has been loaded.")
            return None

        tf = None
        try:
            tf = tarfile.open(self.archive_path, 'r')
            filenames = tf.getnames()
            if filename in filenames:
                to_file = os.path.join(to_dir, filename)
                rf = tf.extractfile(filename)
                with open(to_file, 'w') as wf:
                    xml = rf.read()
                    wf.write(xml)
            else:
                log_info("%s is not found in %s" % (filename, self.archive_path))
                to_file = None
        except tarfile.TarError, e:
            log_error("Failed to read an archive file. file=%s, error=%s" % (self.archive_path, str(e)))
            raise
        except EnvironmentError, e:
            log_error("Failed to write a file. file=%s, error=%s" % (to_file, str(e)))
            raise
        finally:
            if tf is not None:
                tf.close()

        return to_file

    def load(self, archive):
        self.archive_path = archive
        try:
            tf = None
            # Open tar archvie
            tf = tarfile.open(archive, 'r')
            filenames = tf.getnames()
            log_debug("Contents of archive: %s" % filenames)

            # Get file systems
            if FSCONF in filenames:
                f = tf.extractfile(FSCONF)
                fsxml = f.read()
                self.fslist = HDIConfig.parse_filesystems(xml_data=fsxml)

            # Get history count
            if HISTCONF in filenames:
                f = tf.extractfile(HISTCONF)
                reader = csv.reader(f)
                for els in reader:
                    fsname = els[0]
                    count = els[1]
                    self.history_count[fsname] = int(count)

            # Get synced file system information
            for filename in filenames:
                info = backup_sync_common.get_node_and_fsname_from_fsinfo_file(os.path.basename(filename))
                if info is None:
                    # The file is not a synced file system information
                    continue
                ipaddr, fsname = info
                if ipaddr not in self.synced_fslist:
                    self.synced_fslist[ipaddr] = {}
                f = tf.extractfile(filename)
                fsxml = f.read()
                self.synced_fslist[ipaddr][fsname] = HDIConfig.parse_filesystem(xml_data=fsxml)

            fsnames = [e.fileSystemName for e in self.fslist]

            # Get LFT setting
            if LFTCONF in filenames:
                f = tf.extractfile(LFTCONF)
                lftxml = f.read()
                lftinfolist = HDIConfig.parse_arclftctl(xml_data=lftxml)
                for lftinfo in lftinfolist:
                    tmplist = [fsinfo for fsinfo in self.fslist if fsinfo.fileSystemName == lftinfo.fileSystemName]
                    if len(tmplist) == 0:
                        log_warning("File system information and arclftctl result doesn't match. %s doesn't exist in file system information though it exists in arclftctl result." % lftinfo.fileSystemName)
                        # Skip this file system
                        continue
                    fsinfo = tmplist[0]
                    fsinfo.LargeFileTransferSetting = lftinfo

            # Get CIFS shares
            for fsname in fsnames:
                cifsfile = CIFSCONF % fsname
                if cifsfile in filenames:
                    f = tf.extractfile(cifsfile)
                    cifsxml = f.read()
                    self.cifslist.extend(HDIConfig.parse_cifsshares(xml_data=cifsxml))

            # Get NFS shares
            for fsname in fsnames:
                nfsfile = NFSCONF % fsname
                if nfsfile in filenames:
                    f = tf.extractfile(nfsfile)
                    nfsxml = f.read()
                    self.nfslist.extend(HDIConfig.parse_nfsshares(xml_data=nfsxml))

            # Get migration tasks
            if MIGTASKCONF in filenames:
                f = tf.extractfile(MIGTASKCONF)
                taskxml = f.read()
                self.tasklist = HDIConfig.parse_migtasks(xml_data=taskxml)

            # Get cache resident setting
            if RESIDENTCONF in filenames:
                f = tf.extractfile(RESIDENTCONF)
                resxml = f.read()
                self.residentreport = HDIConfig.parse_resident(xml_data=resxml)

            # Get IP addresses
            if NETWORKCONF in filenames:
                f = tf.extractfile(NETWORKCONF)
                ipxml = f.read()
                self.iplist = HDIConfig.parse_interfaces(xml_data=ipxml)

            # Get hostname
            if HOSTCONF in filenames:
                f = tf.extractfile(HOSTCONF)
                self.hostname = f.read().strip()

            # Get UUID
            if UUIDCONF in filenames:
                f = tf.extractfile(UUIDCONF)
                self.uuid = f.read().strip()
        except tarfile.TarError, e:
            log_error("Failed to read an archive file. file=%s, error=%s" % (archive, str(e)))
            raise
        except HDIConfig.FormatError, e:
            log_error("A wrong configuration file is contained in the HDI configuration archive. error=%s, archive=%s" % (str(e), archive))
            raise ConfigurationError("A wrong configuration file was detected in the HDI configuration archive.")
        finally:
            if tf is not None:
                tf.close()

    @staticmethod
    def save_data_to_file(data, to_file):
        log_info("Saving data to %s ..." % to_file)
        try:
            with open(to_file, 'w') as f:
                f.write(data)
        except EnvironmentError, e:
            log_error("Writing data to a file failed. error=%s, file=%s" % (str(e), to_file))
            raise

    @staticmethod
    def rename_file(from_file, to_file):
        log_info("Renaming %s to %s ..." % (from_file, to_file))
        try:
            os.rename(from_file, to_file)
        except EnvironmentError, e:
            log_error("Rename failed. error=%s, from=%s, to=%s" % (str(e), from_file, to_file))
            raise
