#!/usr/bin/python

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "complete_failover / complete_failback"

import sys
import os
import getopt

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import Utility
from Utility import print_status, print_status_warning, print_status_error
import HDIClient
import HDIOperation

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import backup_sync_common
import failover_common
import delete_lft_temporary_objects

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Parse options
    short_opts = ''
    long_opts = ['failback']
    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError, e:
        Exceptions.OptionError(e)

    is_failback = False
    for o, a in opts:
        if o == '--failback':
            is_failback = True

    # Get list of failover jobs
    pending_jobs = sorted(failover_common.get_list_of_jobs(config.fo_conf_dir))
    success_jobs = sorted(failover_common.get_list_of_jobs(config.fo_conf_dir, is_success=True))
    error_jobs = sorted(failover_common.get_list_of_jobs(config.fo_conf_dir, is_error=True))
    log_info("Success: %s, Error: %s, Pending: %s" % (success_jobs, error_jobs, pending_jobs))

    operation = 'failback' if is_failback else 'failover'
    print "%s succeeded file systems: %s" % (operation.capitalize(), ', '.join(success_jobs))
    print "%s failed file systems: %s" % (operation.capitalize(), ', '.join(error_jobs))
    print "%s pending file systems: %s" % (operation.capitalize(), ', '.join(pending_jobs))

    # Confirmation
    answer = Utility.yes_or_no_full("Have you confirmed all configuration has been restored as expected?")
    if answer is False:
        # Canceled
        print_status("Canceled.")
        raise Exceptions.NoExecution()
    if len(error_jobs) != 0 or len(pending_jobs) != 0:
        print_status_warning("There are still error jobs or not-executed jobs!")
    answer = Utility.yes_or_no_full("Are you really OK to complete this %s?" % operation)
    if answer is False:
        # Canceled
        print_status("Canceled.")
        raise Exceptions.NoExecution()

    # Prepare connection to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    # Get HCP information
    print_status("Loading HCP configuration ...")
    try:
        hcpop = HDIOperation.HCPSetting(conn)
        hcp_info = hcpop.get()
    except HDIOperation.OperationError, e:
        log_error("Failed to get HCP information on the node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get HCP information on the node. error=%s" % e.reason, e.code)

    fsop = HDIOperation.FileSystem(conn)
    taskop = HDIOperation.MigrationTask(conn)
    # "Complete" each success failover job
    for name in success_jobs:
        print_status("Executing completion process for %s ..." % name)

        # Get file system information from local node
        try:
            fsinfo = fsop.get(name)
        except HDIOperation.OperationError, e:
            print_status_error("Failed to get file system information of %s. error=%s" % (name, str(e)))
            print_status("Go to the next job ...")
            continue

        # Delete temporary objects on namespace if LFT is enabled
        if fsinfo.LargeFileTransferSetting is not None and fsinfo.LargeFileTransferSetting.largeFileTransfer == 'Enable':
            print_status("Large File Transfer is enabled on %s." % fsinfo.fileSystemName)
            print_status("Cleaning up temporary objects of Large File Transfer ...")
            try:
                delete_lft_temporary_objects.delete_lft_temporary_objects(fsinfo.NamespaceShareSettings.hcpNamespace, hcp_info, config)
            except Exception, e:
                log_error("An error occurred while deleting temporary objects of Large File Transfer function. error=%s" % str(e))
                print_status_error("Skip completion process for %s due to an error." % fsinfo.fileSystemName)
                continue

            # Enable migration task
            print_status("Enabling migration tasks on %s ..." % fsinfo.fileSystemName)
            try:
                tasklist = taskop.get_all(fsname=fsinfo.fileSystemName)
            except HDIOperation.OperationError, e:
                log_error("Failed to get list of migration tasks on %s. error=%s" % (fsinfo.fileSystemName, str(e)))
                print_status_error("Could not enable migration tasks on %s. Please enable them manually." % fsinfo.fileSystemName)
                continue
            for taskinfo in tasklist:
                print_status("Enabling migration task %s ..." % taskinfo.task_name)
                try:
                    taskop.enable(taskinfo.task_name, taskinfo.fileSystemName)
                except HDIOperation.OperationError, e:
                    log_error("Failed to enable task %s. error=%s" % (taskinfo.task_name, str(e)))
                    print_status_error("Failed to enable migration task %s for some reason. Please enable the task manually.")
                    # Keep continue for other tasks
        else:
            print_status("Large File Transfer is not enabled on %s." % fsinfo.fileSystemName)

        # Complete the job
        failover_common.delete_job_file(name, config.fo_conf_dir, is_success=True)
        print_status("%s completed." % name)

    # "Complete" each error failover job
    for name in error_jobs:
        print_status("Executing completion process for %s ..." % name)

        # Do not delete temporary objects even if LFT is enabled because the
        # failover/failback job failed.

        # Complete the job
        failover_common.delete_job_file(name, config.fo_conf_dir, is_error=True)
        print_status("%s completed." % name)

    # "Complete" each pending failover job
    for name in pending_jobs:
        print_status("Executing completion process for %s ..." % name)

        # Complete the job
        failover_common.delete_job_file(name, config.fo_conf_dir)
        print_status("%s completed." % name)

    # Display alert regarding LFT temporary object deletion
    if len(error_jobs) != 0:
        print_status_warning("Did not delete temporary objects of Large File Transfer function for file systems which failed to %s. Please manualy delete them if the failover was actually succeeded and you want to operate the file systems on this node." % operation)

    # Enable backup_configs on this node if it's disabled
    print_status("Enabling execution of backup_configs on this node ...")
    backup_sync_common.enable_backup(config.fo_conf_dir)

    print_status("Completed.")

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        sys.exit(0)
    except Exceptions.OptionError, e:
        log_error("Option error: %s" % str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
