#!/usr/bin/python

from __future__ import with_statement
import os
import errno
import csv
import re

import Utility

try:
    from CommonLogger import log_debug, log_info, log_warning, log_error
except ImportError:
    # To make this module handy and portable
    def log_debug(msg): pass
    def log_info(msg): pass
    def log_warning(msg): print msg
    def log_error(msg): print msg

class BaseError(Exception):
    def __init__(self, reason, code=None):
        self.reason = reason.strip()
        self.code = code

    def __str__(self):
        if self.code is None:
            return "%s" % self.reason
        else:
            return "%s (code=%s)" % (self.reason, self.code)

class FormatError(BaseError):
    pass

class ConfigurationError(BaseError):
    pass

class ParameterError(BaseError):
    pass

class Pair(object):
    FSALL = "<ALL>"
    FSNONE = "<NONE>"

    def __init__(self, info_list=None, primary=None, secondary=None, target_fslist=None, exclude_fslist=None, is_all=False):
        # info_list is a list of values that is recorded in a pair configuration file
        if info_list is None:
            self.primary = primary
            self.secondary = secondary
            self.target_fslist = target_fslist
            self.exclude_fslist = exclude_fslist
            self.is_all = is_all
        else:
            primary = info_list[0]
            secondary = info_list[1]
            fslist = info_list[2:]
            if fslist[0] == Pair.FSNONE:
                target_fslist = exclude_fslist = None
                is_all = False
            elif fslist[0] == Pair.FSALL:
                target_fslist = None
                exclude_fslist = self.get_exclude_fslist(fslist[1:])
                is_all = True
            else:
                target_fslist = self.get_target_fslist(fslist)
                exclude_fslist = self.get_exclude_fslist(fslist)
                is_all = False
            self.primary = primary
            self.secondary = secondary
            self.target_fslist = target_fslist
            self.exclude_fslist = exclude_fslist
            self.is_all = is_all

    def __str__(self):
        return "primary: %s, secondary: %s, target_fslist: %s, exclude_fslist: %s, is_all: %s" % (self.primary, self.secondary, self.target_fslist, self.exclude_fslist, self.is_all)

    def to_list(self):
        result = [self.primary, self.secondary]
        if self.target_fslist is None and self.is_all is False:
            result.extend([Pair.FSNONE])
        elif self.is_all is True:
            result.extend([Pair.FSALL])
            if self.exclude_fslist is not None:
                result.extend(['!' + x for x in self.exclude_fslist])
        else:
            result.extend(self.target_fslist)
            if self.exclude_fslist is not None:
                result.extend(['!' + x for x in self.exclude_fslist])
        log_debug("Pair.to_list: %s" % result)
        return result

    @staticmethod
    def get_target_fslist(fslist):
        # Ex.) ["fs02","fs03","!fs01","!fsRWCS"] -> ["fs02,"fs03"]
        target_fslist = []
        for fs in fslist:
            if not fs.startswith('!'):
                target_fslist.append(fs)
        return target_fslist

    @staticmethod
    def get_exclude_fslist(fslist):
        # Ex.) ["fs02","fs03","!fs01","!fsRWCS"] -> ["fs01,"fsRWCS"]
        exclude_fslist = []
        for fs in fslist:
            if fs.startswith('!'):
                exclude_fslist.append(fs.strip('!'))
        return exclude_fslist

class PairConfig(object):
    def __init__(self, config_dir, confgroup):
        self.config_dir = config_dir
        self.config_file = None
        self.pair_dict = {}
        self.sshkey_dict = {}
        self.max_pair_id = -1
        self.confgroup = confgroup

        # Read config file of the specified configuration group
        self.config_file = os.path.join(self.config_dir, "pair_%s.csv" % confgroup)
        if not os.path.exists(self.config_file):
            # Create a new config file if a new configuration group is specified
            log_info("Pair configuration of the specified configuration group does not exist. Create a new pair configuration. group=%s" % confgroup)
            try:
                with open(self.config_file, 'w'):
                    pass
            except EnvironmentError, e:
                log_error("Failed to create an empty pair configuration file. path=%s, error=%s" % (self.config_file, str(e)))
                raise
            return

        log_info("Reading pair configuration file. file=%s" % self.config_file)
        try:
            with open(self.config_file, 'r') as f:
                reader = csv.reader(f)
                for line in reader:
                    log_debug("Pair config line: %s" % line)
                    if len(line) == 0:
                        log_debug("An empty line. Skip.")
                        continue
                    pid = int(line[0])
                    pair = Pair(info_list=line[3:])
                    self.pair_dict[pid] = pair
                    self.sshkey_dict[pair.primary] = Utility.str2bool(line[1])
                    self.sshkey_dict[pair.secondary] = Utility.str2bool(line[2])
            if len(self.pair_dict) != 0:
                # Not an empty configuration file
                self.max_pair_id = sorted(self.pair_dict.keys(), reverse=True)[0]
        except EnvironmentError, e:
            log_error("An error occurred while opening/reading pair configuration file. file=%s, error=%s" % (self.config_file, str(e)))
            raise
        except (ValueError, IndexError), e:
            log_error("An invalid value is recorded in the pair configuration file. file=%s, error=%s" % (self.config_file, str(e)))
            raise FormatError("An invalid value is recorded in the pair configuration file.")

        for pid in self.pair_dict:
            log_debug("Pair list: %s: %s" % (pid, self.pair_dict[pid]))

    def exists(self, pair_id=None, primary=None, secondary=None):
        if pair_id is None and primary is None and secondary is None:
            log_error("Nothing was specified for PairConfig.exists().")
            raise ParameterError("Nothing was specified for PairConfig.exists().")

        log_debug("Checking pair existence. pair_id=%s, primary=%s, secondary=%s, pair_dict=%s" % (pair_id, primary, secondary, Utility.deep_str(self.pair_dict)))

        result = None
        if pair_id is not None:
            result = pair_id in self.pair_dict
        else:
            for pair in self.pair_dict.values():
                if primary is not None:
                    if secondary is not None:
                        if primary == pair.primary and secondary == pair.secondary:
                            result = True
                            break
                    else:
                        if primary == pair.primary:
                            result = True
                            break
                else:
                    if secondary == pair.secondary:
                        result = True
                        break
            else:
                result = False

        log_debug("Existence: %s" % result)
        return result

    def add(self, primary, secondary, target_fslist=None, exclude_fslist=None, is_all=False):
        """ Add the specified pair to the list.
        Returns pair ID of the new pair.
        """
        log_info("Adding a pair. Primary=%s, secondary=%s" % (primary, secondary))

        # Raise an error if the pair is already registered
        if self.exists((primary, secondary)):
            log_warning("Trying to add a pair that already exists. primary=%s, secondary=%s" % (primary, secondary))
            raise ConfigurationError("The specified pair already exists. primary=%s, secondary=%s" % (primary, secondary))

        # Assign a new pair ID and add to the dictionary
        pid = self.max_pair_id + 1
        self.pair_dict[pid] = Pair(primary=primary, secondary=secondary, target_fslist=target_fslist, exclude_fslist=exclude_fslist, is_all=is_all)

        # SSH key status initialization
        if not primary in self.sshkey_dict:
            self.sshkey_dict[primary] = False
        if not secondary in self.sshkey_dict:
            self.sshkey_dict[secondary] = False

        # Update max pair ID
        self.max_pair_id = pid

        log_debug("Pair added: %s" % self.pair_dict[pid])
        log_debug("Max pair id: %s" % self.max_pair_id)

        return pid

    def edit(self, pair_id, target_fslist=None, exclude_fslist=None, is_all=False):
        """ Edit target file systems of the pair specified by pair_id.
        """
        # Create a new Pair() and overwrite with the same pair ID
        if not self.exists(pair_id):
            raise ParameterError("A wrong pair ID is specified for editing.")

        current_pair = self.pair_dict[pair_id]
        primary = current_pair.primary
        secondary = current_pair.secondary
        self.pair_dict[pair_id] = Pair(primary=primary, secondary=secondary, target_fslist=target_fslist, exclude_fslist=exclude_fslist, is_all=is_all)

    def delete(self, pair_id=None, primary=None, secondary=None):
        """ Delete a pair from pair configuration.

        A pair to be deleted is specified by pair ID or a combination of
        primary node and secondary node. Pair ID has priority than a combination
        of primary and secondary. Primary and secondary must be specified
        if pair ID is not specified.

        This function raises an exception if it found duplicated pair definition.

        Args:
            pair_id (int): Pair ID of the pair to be deleted.
            primary (str): IP address of the primary node of the pair to be deleted.
            secondary (str): IP address of the secondary node of the pair to be
                deleted.
        Returns:
            None
        """
        deleted_pair = None
        if pair_id is not None:
            log_info("Deleting a pair. Pair ID=%s" % pair_id)
            if pair_id in self.pair_dict:
                deleted_pair = self.pair_dict[pair_id]
                del self.pair_dict[pair_id]
            else:
                log_warning("Trying to delete a pair that doesn't exist.")
                # Do nothing if the pair ID is already deleted or does not exist
                return
        elif primary is not None and secondary is not None:
            log_info("Deleting a pair. Primary=%s, secondary=%s" % (primary, secondary))
            to_be_deleted = [pid for pid in self.pair_dict if self.pair_dict[pid].primary == primary and self.pair_dict[pid].secondary == secondary]
            if len(to_be_deleted) > 1:
                # len(to_be_deleted) has to be 0 or 1
                log_error("There are multiple pairs found that has primary=%s and secondary=%s. count=%s" % (primary, secondary, len(to_be_deleted)))
                raise ConfigurationError("There are multiple pairs that have the same primary node and secondary node.")
            elif len(to_be_deleted) == 0:
                log_warning("Trying to delete a pair that doesn't exist.")
                # Finish here. Nothing to do.
                return

            for pid in to_be_deleted:
                deleted_pair = self.pair_dict[pair_id]
                del self.pair_dict[pid]
        else:
            log_error("Nothing was specified or only primary or secondary was specified for PairConfig.delete().")
            raise ParameterError("Too few parameters for PairConfig.delete().")

        # Delete SSH key status
        if deleted_pair is not None:
            for node in (deleted_pair.primary, deleted_pair.secondary):
                if not self.exists(primary=node) and not self.exists(secondary=node):
                    # Delete SSH key status for the node
                    del self.sshkey_dict[node]

    def get_list(self, pair_id=None, primary=None, secondary=None):
        """ Get a list of pairs that meet the specified condition.

        This function returns all pairs if nothing is specified.
        If pair ID is specified, this function returns a list that only contains
        the pair of the pair ID.
        If primary or secondary (or both) is specified, this function returns a list
        of pairs that have the specified primary or secondary (or both).

        If no pairs found for the specified condition, an empty list is returned.

        Args:
            pair_id (int): Pair ID of the pair to get.
            primary (str): IP address of a primary node that you are looking for.
            secondary (str): IP address of a secondary node that you are looking for.
        Returns:
            List of tuple: List of (pair ID, pair)
        """
        pair_list = []
        if pair_id is not None:
            if pair_id in self.pair_dict:
                pair_list = [(pair_id, self.pair_dict[pair_id])]
            else:
                pair_list = []
        elif primary is None and secondary is None:
            # Return all pairs with pair ID order
            pid_list = sorted(self.pair_dict.keys())
            pair_list = []
            for pid in pid_list:
                pair_list.append((pid, self.pair_dict[pid]))
        else:
            log_info("Looking for pairs where primary is %s and secondary is %s" % (primary, secondary))
            if primary is not None and secondary is not None:
                pair_list = [(pid, pair) for pid, pair in self.pair_dict.items() if pair.primary == primary and pair.secondary == secondary]
            elif primary is not None and secondary is None:
                pair_list = [(pid, pair) for pid, pair in self.pair_dict.items() if pair.primary == primary]
            elif primary is None and secondary is not None:
                pair_list = [(pid, pair) for pid, pair in self.pair_dict.items() if pair.secondary == secondary]
            # Sort the list in pair ID order
            pair_list.sort(key=lambda x: x[0])

        log_debug("Pair list: Requested: pair_id=%s Result: %s" % (pair_id, Utility.deep_str(pair_list)))
        return pair_list

    def get_node_list(self):
        """ Returns a list of nodes (IP address) which are included in
        pair configuration.

        Args:
            None
        Returns:
            List(str): List of IP address of nodes that are contained in this
                pair configuration.
        """
        # Get all pairs
        pair_list = self.get_list()

        # Create a list of all nodes
        node_list = [pair.primary for pid, pair in pair_list]
        node_list.extend([pair.secondary for pid, pair in pair_list])

        # Remove duplicate nodes
        node_set = set(node_list)

        log_debug("Node list: %s" % list(node_set))
        return list(node_set)

    def is_ssh_setup(self, node):
        """ Check if Unique SSH key is setup on a node.

        Setup status of Unique SSH key is managed by sshkey_dict.
        This function checks sshkey_dict and returns the status.

        Args:
            node (str): IP address of a node.
        Returns:
            bool: True if Unique SSH is already setup on the node. Otherwise False.
        """

        if node in self.sshkey_dict:
            return self.sshkey_dict[node]
        else:
            # The caller might specify a node that is not defined in pair
            # configuration. This can happen when trying to connect a node
            # which is just being configured for a new pair.
            return False

    def done_ssh_setup(self, node):
        """ Mark a node as 'SSH setup done'.

        Args:
            node (str): IP address of a node.
        Returns:
            None
        """

        log_debug("Marking %s as SSH setup done." % node)
        if node in self.sshkey_dict:
            self.sshkey_dict[node] = True
        else:
            # The caller might specify a node that is not defined in pair
            # configuration. Raise an exception.
            log_error("Unknown node is specified to change SSH setup status. node=%s, SSH status=%s" % (node, self.sshkey_dict))
            raise ParameterError("Unknown node is specified to change SSH status.")

    def save(self):
        log_info("Saving pair configuration to %s." % self.config_file)
        try:
            with open(self.config_file, 'w') as f:
                writer = csv.writer(f, lineterminator='\n')
                sorted_pid_list = sorted(self.pair_dict.keys())
                for pid in sorted_pid_list:
                    pair = self.pair_dict[pid]
                    output = [pid]
                    primary_ssh = Utility.bool2str(self.is_ssh_setup(pair.primary))
                    secondary_ssh = Utility.bool2str(self.is_ssh_setup(pair.secondary))
                    output.extend([primary_ssh, secondary_ssh])
                    output.extend(pair.to_list())
                    writer.writerow(output)
        except EnvironmentError, e:
            log_error("An error occurred while saving pair configuration file. file=%s, error=%s" % (self.config_file, str(e)))
            raise

    @classmethod
    def find_matched_ip(cls, config_dir, iplist, confgroup=None):
        log_info("List of IP address: %s" % iplist)
        if confgroup is None:
            confgroup_list = list_configuration_group(config_dir)
            for confgroup in confgroup_list:
                pairconf = PairConfig(config_dir, confgroup)
                node_list = pairconf.get_node_list()
                log_info("List of nodes (%s): %s" % (confgroup, node_list))
                node_set = set(node_list)
                ip_set = set(iplist)
                myself = list(node_set & ip_set)
                if len(myself) != 0:
                    return myself[0]

            # Reach here if no return happened in the loop above
            return None
        else:
            pairconf = PairConfig(config_dir, confgroup)
            node_list = pairconf.get_node_list()
            log_info("List of nodes (%s): %s" % (confgroup, node_list))
            node_set = set(node_list)
            ip_set = set(iplist)
            myself = list(node_set & ip_set)
            if len(myself) != 0:
                return myself[0]
            else:
                return None

class PairGroup(object):
    def __init__(self, pair_list=None):
        # List of pair ID
        self.pair_list = pair_list

    def __str__(self):
        return "%s" % self.pair_list

class PairGroupConfig(object):
    def __init__(self, config_dir, confgroup):
        self.config_dir = config_dir
        self.config_file = None
        self.confgroup = confgroup
        self.pair_group_dict = {}

        # Read config file of the specified configuration group
        self.config_file = os.path.join(self.config_dir, "pairgroup_%s.csv" % confgroup)
        if not os.path.exists(self.config_file):
            # Create a new config file if a new configuration group is specified
            log_info("Pair group configuration of the specified configuration group does not exist. Create a new pair group configuration. group=%s" % confgroup)
            try:
                with open(self.config_file, 'w'):
                    pass
            except EnvironmentError, e:
                log_error("Failed to create an empty pair group configuration file. path=%s, error=%s" % (self.config_file, str(e)))
                raise
            return

        log_info("Reading pair group configuration file. file=%s" % self.config_file)
        try:
            with open(self.config_file, 'r') as f:
                reader = csv.reader(f)
                for line in reader:
                    log_debug("Pair group config line: %s" % line)
                    if len(line) == 0:
                        log_debug("An empty line. Skip.")
                        continue
                    name = line[0]
                    if name in self.pair_group_dict:
                        log_warning("Duplicate pair group name found. Overwrite the existing pair group. name=%s" % name)
                    self.pair_group_dict[name] = PairGroup(map(int, line[1:]))
        except EnvironmentError, e:
            log_error("An error occurred while opening/reading pair group configuration file. file=%s, error=%s" % (self.config_file, str(e)))
            raise
        except (ValueError, IndexError), e:
            log_error("An invalid value is recorded in the pair group configuration file. file=%s, error=%s" % (self.config_file, str(e)))
            raise FormatError("An invalid value is recorded in the pair group configuration file.")

    def exists(self, name):
        return name in self.pair_group_dict

    def add(self, name, pair_list):
        """ Add a pair group.

        Create a pair group with a specified name and pair ID list, and
        register it to myself.

        Args:
            name (str): Name of a new pair group.
            pair_list (list of int): List of pair ID for the pair group.
        Returns:
            None
        """
        log_info("Adding a pair group. name=%s, pair_list=%s" % (name, pair_list))
        if name in self.pair_group_dict:
            log_warning("The specified name is already used for pair group name. name=%s" % name)
            raise ConfigurationError("The specified pair group already exists. name=%s" % name)
        self.pair_group_dict[name] = PairGroup(pair_list)

    def delete(self, name):
        log_info("Deleting a pair group. name=%s" % name)
        if name in self.pair_group_dict:
            del self.pair_group_dict[name]
        else:
            log_warning("Trying to delete a pair group that does not exist.")
            # Do nothing if the pair group is already deleted or does not exist

    def update(self, name, pair_list):
        log_info("Updating a pair group: name=%s, pair_list=%s" % (name, pair_list))
        if not name in self.pair_group_dict:
            log_warning("The specified has not been added yet. Forcibly adding the pair group. name=%s" % name)
        self.pair_group_dict[name] = PairGroup(pair_list)

    def get_list(self, name=None):
        # Returns a list of (pair group name, PairGroup)
        pair_group_list = []
        if name is not None:
            # Returns a pair group of the specified name
            if name in self.pair_group_dict:
                pair_group_list = [(name, self.pair_group_dict[name])]
            else:
                pair_group_list = []
        else:
            # Returns all pair groups with name order
            sorted_name_list = sorted(self.pair_group_dict.keys())
            pair_group_list = []
            for name in sorted_name_list:
                pair_group_list.append((name, self.pair_group_dict[name]))

        log_debug("PairGroup list: Requested: name=%s Result: %s" % (name, Utility.deep_str(pair_group_list)))
        return pair_group_list

    def save(self):
        log_info("Saving pair group configuration to %s." % self.config_file)
        try:
            with open(self.config_file, 'w') as f:
                writer = csv.writer(f, lineterminator='\n')
                sorted_name_list = sorted(self.pair_group_dict.keys())
                for name in sorted_name_list:
                    pair_group = self.pair_group_dict[name]
                    output = [name]
                    output.extend(pair_group.pair_list)
                    writer.writerow(output)
        except EnvironmentError, e:
            log_error("An error occurred while saving pair group configuration file. file=%s, error=%s" % (self.config_file, str(e)))
            raise

class StoragePolicy(object):
    AUTOTOTAL = 0
    AUTOUSAGE = 1
    AUTOSPEC = 2
    AUTOPICKUP = 3
    LIKEFORLIKE = 4
    MANUAL = 5
    FAILBACK = 6

    FSDEF = "<DEF>"

    def __init__(self, target_fs, policy_string=None, policy=None, margin=None, spec=None, fslulist=None, wslulist=None, fs_capacity=None, ws_capacity=None):
        self.target_fs = target_fs

        # Initialize other parameters with None
        self.policy = None
        self.margin_capacity = None
        self.spec_capacity = None
        self.fslulist = None
        self.wslulist = None
        self.fs_capacity = None
        self.ws_capacity = None

        if policy_string is not None:
            # Configure the policy from policy_string
            policy = policy_string.split(':')[0]
            if policy == "autototal":
                self.policy = StoragePolicy.AUTOTOTAL
            elif policy == "autousage":
                self.policy = StoragePolicy.AUTOUSAGE
            elif policy == "autospec":
                self.policy = StoragePolicy.AUTOSPEC
            elif policy == "autopickup":
                self.policy = StoragePolicy.AUTOPICKUP
            elif policy == "likeforlike":
                self.policy = StoragePolicy.LIKEFORLIKE
            elif policy == "manual":
                self.policy = StoragePolicy.MANUAL
            elif policy == "failback":
                self.policy = StoragePolicy.FAILBACK
            else:
                log_error("A wrong string for policy was provided. policy_string=%s" % policy_string)
                raise FormatError("A wrong string for policy was provided. policy_string=%s" % policy_string)

            try:
                if self.policy == StoragePolicy.AUTOUSAGE:
                    # Ex.) "autousage:50"
                    self.margin_capacity = int(policy_string.split(':')[1])
                elif self.policy == StoragePolicy.AUTOSPEC:
                    # Ex.) "autospec:100"
                    self.spec_capacity = int(policy_string.split(':')[1])
                elif self.policy == StoragePolicy.MANUAL:
                    fsstorage = policy_string.split(':')[1]
                    wsstorage = policy_string.split(':')[2]
                    if fsstorage.startswith("lu"):
                        # Cluster
                        # Ex.) "manual:lu00-lu01-lu02:lu03"
                        self.fslulist = fsstorage.split('-')
                        if len(wsstorage) == 0:
                            # If wsstorage is an empty string, wsstorage.split('-')
                            # will result in a list with a single element ''.
                            self.wslulist = []
                        else:
                            self.wslulist = wsstorage.split('-')

                        # Sanity check
                        if len(self.fslulist) == 0:
                            # self.wslulist can be empty if AFM is not used
                            log_error("No LUs are defined for 'manual' policy. policy_string=%s" % policy_string)
                            raise FormatError("No LUs are defined for 'manual' policy.")

                        if sum([x.startswith("lu") for x in self.fslulist]) != len(self.fslulist):
                            log_error("LU information for 'manual' policy does not start with 'lu'. policy_string=%s, fslulist=%s" % (policy_string, self.fslulist))
                            raise FormatError("LU name does not start with 'lu' in provided storage configuration policy.")
                        if sum([x.startswith("lu") for x in self.wslulist]) != len(self.wslulist):
                            log_error("LU information for 'manual' policy does not start with 'lu'. policy_string=%s, fslulist=%s" % (policy_string, self.fslulist))
                            raise FormatError("LU name does not start with 'lu' in provided storage configuration policy.")
                    else:
                        # Single
                        # Ex.) "manual:100:30" or "manual:300:auto"
                        self.fs_capacity = int(fsstorage)   # ValueError will be raised if fsstorage is a wrong value
                        if wsstorage.isdigit():
                            self.ws_capacity = int(wsstorage)
                        elif wsstorage == "auto":
                            self.ws_capacity = None
                        else:
                            log_warning("Unknown value is specified for workspace capacity in storage configuration policy. value=%s" % wsstorage)
                            self.ws_capacity = None
            except (ValueError, IndexError), e:
                log_error("Wrong format was detected in storage policy configuration file. error=%s, policy_string=%s" % (str(e), policy_string))
                raise FormatError("Wrong format was detected in storage policy configuration file.")
        else:
            # Configure the policy from parameters passed to the constructor
            if policy is None or policy not in (StoragePolicy.AUTOTOTAL, StoragePolicy.AUTOUSAGE, StoragePolicy.AUTOSPEC, StoragePolicy.AUTOPICKUP, StoragePolicy.LIKEFORLIKE, StoragePolicy.MANUAL, StoragePolicy.FAILBACK):
                log_error("Wrong policy is specified for StoragePolicy. policy=%s" % policy)
                raise ParameterError("Wrong policy is specified for StoragePolicy. policy=%s" % policy)

            self.policy = policy
            if policy == StoragePolicy.AUTOUSAGE:
                if margin is None:
                    log_error("Margin capacity is not specified for 'Auto with usage' policy.")
                    raise ParameterError("Margin capacity is not specified for 'Auto with usage' policy.")
                self.margin_capacity = margin
            elif policy == StoragePolicy.AUTOSPEC:
                if spec is None:
                    log_error("Margin capacity is not specified for 'Auto with specified' policy.")
                    raise ParameterError("Margin capacity is not specified for 'Auto with specified policy.")
                self.spec_capacity = spec
            elif policy == StoragePolicy.MANUAL:
                self.fslulist = fslulist
                self.wslulist = wslulist
                self.fs_capacity = fs_capacity
                self.ws_capacity = ws_capacity

    def to_string(self):
        if self.policy == StoragePolicy.AUTOTOTAL:
            s = "autototal"
        elif self.policy == StoragePolicy.AUTOUSAGE:
            s = "autousage:%d" % self.margin_capacity
        elif self.policy == StoragePolicy.AUTOSPEC:
            s = "autospec:%d" % self.spec_capacity
        elif self.policy == StoragePolicy.AUTOPICKUP:
            s = "autopickup"
        elif self.policy == StoragePolicy.LIKEFORLIKE:
            s = "likeforlike"
        elif self.policy == StoragePolicy.MANUAL:
            if self.fslulist is not None:
                # Cluster
                s = "manual:%s:%s" % ('-'.join(self.fslulist), '-'.join(self.wslulist))
            else:
                # Single
                if self.ws_capacity is not None:
                    s = "manual:%d:%d" % (self.fs_capacity, self.ws_capacity)
                else:
                    s = "manual:%d:%s" % (self.fs_capacity, "auto")
        elif self.policy == StoragePolicy.FAILBACK:
            s = "failback"
        return s

    def to_display_string(self):
        if self.policy == StoragePolicy.AUTOTOTAL:
            s = StoragePolicy.policy_to_display_string(self.policy)
        elif self.policy == StoragePolicy.AUTOUSAGE:
            s = StoragePolicy.policy_to_display_string(self.policy) + " (with %s GB margin)" % self.margin_capacity
        elif self.policy == StoragePolicy.AUTOSPEC:
            s = StoragePolicy.policy_to_display_string(self.policy) + " (%s GB)" % self.spec_capacity
        elif self.policy == StoragePolicy.AUTOPICKUP:
            s = StoragePolicy.policy_to_display_string(self.policy)
        elif self.policy == StoragePolicy.LIKEFORLIKE:
            s = StoragePolicy.policy_to_display_string(self.policy)
        elif self.policy == StoragePolicy.MANUAL:
            if self.fslulist is not None:
                # Cluster
                s = StoragePolicy.policy_to_display_string(self.policy) + " - FS: %s WS: %s" % (','.join(self.fslulist), ','.join(self.wslulist))
            else:
                # Single
                if self.ws_capacity is not None:
                    s = StoragePolicy.policy_to_display_string(self.policy) + " - FS: %s GB WS: %s GB" % (self.fs_capacity, self.ws_capacity)
                else:
                    s = StoragePolicy.policy_to_display_string(self.policy) + " - FS: %s GB WS: %s" % (self.fs_capacity, "auto")
        elif self.policy == StoragePolicy.FAILBACK:
            s = StoragePolicy.policy_to_display_string(self.policy)

        return s

    @classmethod
    def policy_to_display_string(cls, policy):
        """ Return a string description of the specified policy constant.
        for display.
        """

        if policy == cls.AUTOTOTAL:
            s = "Auto with total"
        elif policy == cls.AUTOUSAGE:
            s = "Auto with usage"
        elif policy == cls.AUTOSPEC:
            s = "Auto with specified"
        elif policy == cls.AUTOPICKUP:
            s = "Auto with pickup one"
        elif policy == cls.LIKEFORLIKE:
            s = "Like for Like"
        elif policy == cls.MANUAL:
            s = "Manual"
        elif policy == cls.FAILBACK:
            s = "Failback"

        return s

    @classmethod
    def policy_to_help_string_cluster(cls, policy):
        """ Return a string description as help explanation of a policy
        for cluster model.
        """

        if policy == cls.AUTOTOTAL:
            msg = """
\tPick up an empty LU(s) in numerical order until the capacity reaches
\tto the capacify of the file system on the original node.
"""
        elif policy == cls.AUTOUSAGE:
            msg = """
\tPick up an empty LU(s) in numerical order until the total capacity reaches
\tto the USED capacity of the file system on the original node.
\tOptionally a 'margin' capacity can be specified which will be additionally
\tallocated to the USED capacity on the original node.
"""
        elif policy == cls.AUTOSPEC:
            msg = """
\tPick up an empty LU(s) in numerical order until the total capacity reaches
\tto a specific capacity.
\tYou have to specify a capacity when selecting this option.
"""
        elif policy == cls.AUTOPICKUP:
            msg = """
\tPick up just one empty LU regardless of its capacity.
"""
        elif policy == cls.MANUAL:
            msg = """
\tPick up a specified LU(s).
"""
        elif policy == cls.LIKEFORLIKE:
            msg = """
\tPick up an LU(s) completely same as the original node.
\tCapacity of each LU is NOT considered. Pick up the same NAME LU(s) as it is
\ton the original node.
"""
        elif policy == cls.FAILBACK:
            msg = """
\tPick up the same LU(s) that was used on the primary node.
\tSelectable only failback case.
"""
        else:
            msg = "\tUnknown policy!"

        return msg

    @classmethod
    def policy_to_help_string_single(cls, policy):
        """ Return a string description as help explanation of a policy
        for single model.
        """

        if policy == cls.AUTOTOTAL:
            msg = """
\tAssign the same capacity with the file system on the original node.
"""
        elif policy == cls.AUTOUSAGE:
            msg = """
\tAssign the USED capacity of the file system on the original node.
\tOptionally a 'margin' capacity can be specified which will be additionally
\tallocated to the USED capacity on the original node.
"""
        elif policy == cls.AUTOSPEC:
            msg = """
\tAssign a specific capacity.
\tYou have to specify a capacity when selecting this option. The same
\tcapacity will be assigned to the file systems that this policy is
\tapplied to.
"""
        elif policy == cls.MANUAL:
            msg = """
\tAssign a specific capacity.
\tYou have to specify a capacity for each file system that this policy
\tis applied to.
"""
        elif policy == cls.LIKEFORLIKE:
            msg = """
\tAssign the same capacity as the file system on the original node for
\teach file system.
"""
        elif policy == cls.FAILBACK:
            msg = """
\tAssign the same capacity that was assigned on the primary node.
\tSelectable only failback case.
"""
        else:
            msg = "\tUnknown policy!"

        return msg

    @classmethod
    def get_available_policies(cls):
        """ Return a list of policy constants.
        """

        # Do not return FAILBACK.
        # FAILBACK is automatically selected if and only if failback is executed.
        return [cls.AUTOTOTAL, cls.AUTOUSAGE, cls.AUTOSPEC, cls.AUTOPICKUP, cls.LIKEFORLIKE, cls.MANUAL]

    def __str__(self):
        return "target_fs: %s, policy: %s" % (self.target_fs, self.to_string())

class StoragePolicyConfig(object):
    '''
    Storage configuration file example:
        PID1,<DEF>,policy1
        PID1,fs2,policy2
        PID2,<DEF>,policy3
        PID3,<DEF>,policy3
        PID3,fs1,policy1
    '''
    def __init__(self, config_dir, confgroup):
        self.config_dir = config_dir
        self.config_file = None
        self.confgroup = confgroup
        # policy_dict is a dictionary that has pair ID as a key and list of StoragePolicy
        # for the pair as a value
        self.policy_dict = {}

        # Read config file of the specified configuration group
        self.config_file = os.path.join(self.config_dir, "storage_%s.csv" % confgroup)
        if not os.path.exists(self.config_file):
            # Create a new config file if a new configuration group is specified
            log_info("Storage policy configuration of the specified configuration group does not exist. Create a new storage policy configuration. group=%s" % confgroup)
            try:
                with open(self.config_file, 'w'):
                    pass
            except EnvironmentError, e:
                log_error("Failed to create an empty storage policy configuration file. path=%s, error=%s" % (self.config_file, str(e)))
                raise
            return

        log_info("Reading storage policy configuration file. file=%s" % self.config_file)
        try:
            with open(self.config_file, 'r') as f:
                reader = csv.reader(f)
                for line in reader:
                    log_debug("Storage policy config line: %s" % line)
                    if len(line) == 0:
                        log_debug("An empty line. Skip.")
                        continue
                    pair_id = int(line[0])
                    target_fs = line[1]
                    policy_string = line[2]
                    if pair_id not in self.policy_dict:
                        self.policy_dict[pair_id] = []
                    self.policy_dict[pair_id].append(StoragePolicy(target_fs, policy_string))
        except EnvironmentError, e:
            log_error("An error occurred while reading storage policy configuration file. file=%s, error=%s" % (self.config_file, str(e)))
            raise
        except (ValueError, IndexError), e:
            log_error("An invalid value is recorded in the storage policy configuration file. file=%s, error=%s" % (self.config_file, str(e)))
            raise FormatError("An invalid value is recorded in the storage policy configuration file.")

    def get_list(self, pair_id=None, target_fs=None):
        # Returns a list of (pair id, target_fs, policy)
        policy_list = []
        if pair_id is None:
            # Returns all policies if no pair ID is specified
            sorted_pid_list = sorted(self.policy_dict.keys())
            for pid in sorted_pid_list:
                policy_list_pid = self.policy_dict[pid]
                for policy in policy_list_pid:
                    policy_list.append((pid, policy.target_fs, policy))
        else:
            # Returns the policies for the specified pair
            if not pair_id in self.policy_dict:
                # The specified pair does not exist in policy list
                log_warning("No policies are defined for the requested pair ID. pair_id=%s" % pair_id)
                policy_list = []
            else:
                for policy in self.policy_dict[pair_id]:
                    if target_fs is None or target_fs == policy.target_fs:
                        policy_list.append((pair_id, policy.target_fs, policy))

        log_debug("Policy list: Requested: pair_id=%s, target_fs=%s Result: %s" % (pair_id, target_fs, Utility.deep_str(policy_list)))
        return policy_list

    def get_policy(self, pair_id, target_fs):
        """ Get storage policy for a file system in a pair.

        Get storage policy for the specified file system in the specified
        pair.
        If no specific policy is defined for the file system, get default
        policy instead.

        Args:
            pair_id (int): Pair ID of the pair to get a storage policy.
            target_fs (str): File system name to get a storage policy.
        Returns:
            StoragePolicy: Storage policy of the file system.
                None is returned if neither a storage policy for the file
                system nor default policy is defined, or if the specified
                pair ID does not exist.
        """
        if not pair_id in self.policy_dict:
            # The specified pair does not exist in policy list
            log_warning("No policies are defined for the requested pair ID. pair_id=%s" % pair_id)
            return None
        else:
            policy_list = self.policy_dict[pair_id]
            fs_policy_list = [policy for policy in policy_list if policy.target_fs == target_fs]
            if len(fs_policy_list) == 0:
                policy = self.get_default_policy(pair_id)
            else:
                policy = fs_policy_list[0]
            return policy

    def get_default_policy(self, pair_id):
        """ Get default policy of the specified pair.

        Args:
            pair_id (int): Pair ID of the pair to get default policy.
        Returns:
            StoragePolicy: Default storage policy of the pair. If not found,
                None is returned.
        """
        if not pair_id in self.policy_dict:
            # The specified pair does not exist in policy list
            log_warning("No policies are defined for the requested pair ID. pair_id=%s" % pair_id)
            return None
        else:
            policy_list = self.policy_dict[pair_id]
            def_policy_list = [policy for policy in policy_list if policy.target_fs == StoragePolicy.FSDEF]
            if len(def_policy_list) == 0:
                log_error("No default policy is defined for pair ID %s." % pair_id)
                return None
            return def_policy_list[0]

    def add(self, pair_id, target_fs=None, policy_string=None, policy=None):
        log_info("Adding a storage policy. pair_id=%s, target_fs=%s, policy_string=%s, policy=%s" % (pair_id, target_fs, policy_string, policy))
        # Parameter check
        if policy is None:
            if target_fs is None or policy_string is None:
                raise ParameterError("target_fs and policy_string must be specified if policy is not specified.")
        else:
            if target_fs is not None or policy_string is not None:
                raise ParameterError("target_fs and policy_string must not be specified if policy is specified.")

        if not pair_id in self.policy_dict:
            log_debug("New storage policy is being added for pair ID %s" % pair_id)
            self.policy_dict[pair_id] = []
        else:
            target_fslist = [e.target_fs for e in self.policy_dict[pair_id]]
            if target_fs is None:
                target_fs = policy.target_fs
            if target_fs in target_fslist:
                log_warning("A storage policy for the specified file system in the pair ID already exists. pair_id=%s, target_fs=%s, target_fslist=%s" % (pair_id, target_fs, target_fslist))
                raise ConfigurationError("A storage policy for the specified file system in the pair ID already exists.")
        if policy is None:
            self.policy_dict[pair_id].append(StoragePolicy(target_fs, policy_string))
        else:
            self.policy_dict[pair_id].append(policy)

    def delete(self, pair_id, target_fs=None):
        # Delete policies for the specified pair (only target fs if it is specified)
        log_info("Deleting storage policy for a pair. pair_id=%s, target_fs=%s" % (pair_id, target_fs))
        if not pair_id in self.policy_dict:
            log_warning("Storage policies for the specified pair do not exist. pair_id=%s" % pair_id)
            return
        if target_fs is None:
            del self.policy_dict[pair_id]
        else:
            policy_list_pair = self.policy_dict[pair_id]
            log_debug("Policy list for Pair ID %s: %s" % (pair_id, Utility.deep_str(policy_list_pair)))
            for idx, policy in enumerate(policy_list_pair):
                if policy.target_fs == target_fs:
                    break
            else:
                log_info("No storage policies for the specified file system and pair are defined. pair_id=%s, target_fs=%s" % (pair_id, target_fs))
                idx = None
            if idx is not None:
                del policy_list_pair[idx]

    def save(self):
        log_info("Saving storage policies to %s." % self.config_file)
        try:
            with open(self.config_file, 'w') as f:
                writer = csv.writer(f, lineterminator='\n')
                sorted_pair_id_list = sorted(self.policy_dict.keys())
                for pair_id in sorted_pair_id_list:
                    policy_list_pair = self.policy_dict[pair_id]
                    for policy in policy_list_pair:
                        output = [pair_id, policy.target_fs, policy.to_string()]
                        writer.writerow(output)
        except EnvironmentError, e:
            log_error("An error occurred while saving storage policy configuration file. file=%s, error=%s" % (self.config_file, str(e)))
            raise

# Function

def list_configuration_group(config_dir):
    try:
        items = os.listdir(config_dir)
    except EnvironmentError, e:
        if e.errno == errno.ENOENT:
            log_error("The specified configuration directory does not exist. dir=%s" % config_dir)
            raise ConfigurationError("The specified directory does not exist. dir=%s" % config_dir)
        else:
            log_error("An error occurred while getting a list of files under %s. error=%s" % (config_dir, str(e)))
            raise
    log_debug("Files under %s: %s" % (config_dir, items))

    namelist = []
    for item in items:
        # Use pair configuration files to list up configuration groups.
        m = re.match(r"pair_([a-zA-Z0-9_]+).csv", item)
        if m is not None:
            namelist.append(m.group(1))

    return namelist
