#!/usr/bin/python

from __future__ import with_statement
import xml.etree.ElementTree as ET
from xml.parsers.expat import ExpatError
import re

import Utility

try:
    from CommonLogger import log_debug, log_info, log_warning, log_error
except ImportError:
    # To make this module handy and portable
    def log_debug(msg): pass
    def log_info(msg): pass
    def log_warning(msg): print msg
    def log_error(msg): print msg

# Constants

LFT_VERSION = Utility.Version('6.4.0')
LFT_MINCAPACITY = 100

# Classes

class BaseError(Exception):
    def __init__(self, reason, code=None):
        self.reason = reason.strip()
        self.code = code

    def __str__(self):
        if self.code is None:
            return "%s" % self.reason
        else:
            return "%s (code=%s)" % (self.reason, self.code)

class FormatError(BaseError):
    pass

class ParameterError(BaseError):
    pass

class ConfigurationError(BaseError):
    pass

class UnsupportedError(BaseError):
    pass

class Model(object):
    CLUSTER = 0
    SINGLE = 1

    @classmethod
    def tostr(cls, model):
        if model == cls.CLUSTER:
            return "CLUSTER"
        elif model == cls.SINGLE:
            return "SINGLE"
        else:
            return "UNKNOWN"

class NamespaceInfo(object):
    def __init__(self, nselm=None):
        # nselm is an XML element of 'NamespaceShareSettings'
        self.namespaceType = None
        self.type = None
        self.FQDN = None
        self.externalHcpHostName = None
        self.user = None
        self.password = None
        self.hcpReplicaHost = None
        self.externalReplicaHcpHostName = None
        self.hcpNamespace = None
        self.xml = None

        if nselm is None:
            return

        if nselm.tag != 'NamespaceShareSettings':
            log_error("A wrong XML element was passed for NamespaceInfo. tag=%s" % nselm.tag)
            raise ParameterError("A wrong XML element for NamespaceInfo. tag=%s" % nselm.tag)

        e = nselm.find('namespaceType')
        self.namespaceType = (e.text if e.text != "--" else None) if e is not None else None
        e = nselm.find('type')
        self.type = (e.text if e.text != "--" else None) if e is not None else None
        e = nselm.find('FQDN')
        self.FQDN = e.text if e is not None else None
        e = nselm.find('externalHcpHostName')
        self.externalHcpHostName = e.text if e is not None else None
        e = nselm.find('user')
        self.user = e.text if e is not None else None
        # Doesn't get 'password' element because it is always empty
        e = nselm.find('hcpReplicaHost')
        self.hcpReplicaHost = e.text if e is not None else None
        e = nselm.find('externalReplicaHcpHostName')
        self.externalReplicaHcpHostName = e.text if e is not None else None
        e = nselm.find('hcpNamespace')
        self.hcpNamespace = e.text if e is not None else None

        self.xml = ET.tostring(nselm, encoding='UTF-8')

class WormInfo(object):
    def __init__(self, wormelm=None):
        # wormelm is an XML element of 'WormSetting'
        self.maxRetention = None
        self.minRetention = None
        self.isAutoCommitEnabled = None
        self.commitModeSetting = None
        self.autoCommitPeriod = None
        self.defaultRetention = None
        self.isWormDeleteEnabled = None
        self.longestRetention = None
        self.xml = None

        if wormelm is None:
            return

        if wormelm.tag != 'WormSetting':
            log_error("A wrong XML element was passed for WormInfo. tag=%s" % wormelm.tag)
            raise ParameterError("A wrong XML element for WormInfo. tag=%s" % wormelm.tag)

        e = wormelm.find('maxRetention')
        self.maxRetention = e.text if e is not None else None
        e = wormelm.find('minRetention')
        self.minRetention = e.text if e is not None else None
        e = wormelm.find('isAutoCommitEnabled')
        self.isAutoCommitEnabled = (Utility.str2bool(e.text) if e.text is not None else None) if e is not None else None
        e = wormelm.find('commitModeSetting')
        self.commitModeSetting = e.text if e is not None else None
        e = wormelm.find('autoCommitPeriod')
        self.autoCommitPeriod = e.text if e is not None else None
        e = wormelm.find('defaultRetention')
        self.defaultRetention = e.text if e is not None else None
        e = wormelm.find('isWormDeleteEnabled')
        self.isWormDeleteEnabled = (Utility.str2bool(e.text) if e.text is not None else None) if e is not None else None
        e = wormelm.find('longestRetention')
        self.longestRetention = (int(e.text) if e.text is not None else None) if e is not None else None

        self.xml = ET.tostring(wormelm, encoding='UTF-8')

    def has_different_configuration_from(self, worminfo):
        # Max retention
        if worminfo.maxRetention != self.maxRetention:
            log_info("WORM max retention has been changed. previous:%s, current:%s" % (worminfo.maxRetention, self.maxRetention))
            return True
        # Min retention
        if worminfo.minRetention != self.minRetention:
            log_info("WORM min retention has been changed. previous:%s, current:%s" % (worminfo.minRetention, self.minRetention))
            return True
        # Autocommit usage
        if worminfo.isAutoCommitEnabled != self.isAutoCommitEnabled:
            log_info("WORM Auto Commit usage has been changed. previous:%s, current:%s" % (worminfo.isAutoCommitEnabled, self.isAutoCommitEnabled))
            return True
        # Autocommit mode
        if worminfo.commitModeSetting != self.commitModeSetting:
            log_info("WORM Auto Commit mode has been changed. previous:%s, current:%s" % (worminfo.commitModeSetting, self.commitModeSetting))
            return True
        # Autocommit grace period
        if worminfo.autoCommitPeriod != self.autoCommitPeriod:
            log_info("WORM Auto Commit grace period has been changed. previous:%s, current:%s" % (worminfo.autoCommitPeriod, self.autoCommitPeriod))
            return True
        # Autocommit default retention
        if worminfo.defaultRetention != self.defaultRetention:
            log_info("WORM Auto Commit default retention has been changed. previous:%s, current:%s" % (worminfo.defaultRetention, self.defaultRetention))
            return True

        return False

class MountInfo(object):
    def __init__(self, mountelm=None):
        # mountelm is an XML element of 'MountSetting'
        self.isMounted = None
        self.mountStatus = None
        self.isReadOnly = None
        self.isFileLastAccessTimeRecordingEnabled = None
        self.xml = None

        if mountelm is None:
            return

        if mountelm.tag != 'MountSetting':
            log_error("A wrong XML element was passed for MountInfo. tag=%s" % mountelm.tag)
            raise ParameterError("A wrong XML element for MountInfo. tag=%s" % mountelm.tag)

        e = mountelm.find('isMounted')
        self.isMounted = (Utility.str2bool(e.text) if e.text is not None else None) if e is not None else None
        e = mountelm.find('mountStatus')
        self.mountStatus = e.text if e is not None else None
        e = mountelm.find('isReadOnly')
        self.isReadOnly = (Utility.str2bool(e.text) if e.text is not None else None) if e is not None else None
        e = mountelm.find('isFileLastAccessTimeRecordingEnabled')
        self.isFileLastAccessTimeRecordingEnabled = (Utility.str2bool(e.text) if e.text is not None else None) if e is not None else None

        self.xml = ET.tostring(mountelm, encoding='UTF-8')

    def toxml(self):
        log_info("Building an XML for mount settings.")

        mount = ET.Element("MountSetting")

        # isMounted
        if self.isMounted is None:
            log_warning("Mount status is not defined. 'false' will be set.")
            value = 'false'
        else:
            value = Utility.bool2str(self.isMounted)
        ET.SubElement(mount, "isMounted").text = value

        # isReadOnly
        if self.isReadOnly is None:
            if self.isMounted:
                # isReadOnly property is required if isMounted is set to 'true'
                log_error("A wrong value in mount XML. isReadOnly value is missing while isMounted is true.")
                raise FormatError("isReadOnly value is missing while isMounted is true.")
            else:
                log_warning("Read-only mount status is not defined. This parameter will be omitted.")
        else:
            ET.SubElement(mount, "isReadOnly").text = Utility.bool2str(self.isReadOnly)

        # isFileLastAccessTimeRecordingEnabled
        if self.isFileLastAccessTimeRecordingEnabled is None:
            log_warning("Last file system access time recording is not defined. This parameter will be omitted.")
        else:
            ET.SubElement(mount, "isFileLastAccessTimeRecordingEnabled").text = Utility.bool2str(self.isFileLastAccessTimeRecordingEnabled)

        # Does not support isConvertToAdvancedAclType property. Does not specify it.

        return ET.tostring(mount, encoding='UTF-8')

class FileSystemLUInfo(object):
    def __init__(self, luelm=None):
        # luelm is an XML element of 'LU'
        self.deviceFileName = None
        self.xml = None

        if luelm is None:
            return

        if luelm.tag != 'LU':
            log_error("A wrong XML element was passed for FileSystemLUInfo. tag=%s" % luelm.tag)
            raise ParameterError("A wrong XML element for FileSystemLUInfo. tag=%s" % luelm.tag)

        e = luelm.find('deviceFileName')
        self.deviceFileName = e.text if e is not None else None

        self.xml = ET.tostring(luelm, encoding='UTF-8')

class WorkSpaceLUInfo(object):
    def __init__(self, wselm=None):
        # wselm is an XML element of 'WorkSpace'
        self.deviceFileName = None
        self.xml = None

        if wselm is None:
            return

        if wselm.tag != 'WorkSpace':
            log_error("A wrong XML element was passed for WorkSpaceLUInfo. tag=%s" % wselm.tag)
            raise ParameterError("A wrong XML element for WorkSpaceLUInfo. tag=%s" % wselm.tag)

        e = wselm.find('deviceFileName')
        self.deviceFileName = e.text if e is not None else None

        self.xml = ET.tostring(wselm, encoding='UTF-8')

class LargeFileTransferInfo(object):
    def __init__(self, lftelm=None):
        # lftelm is an XML element of 'LargeFileTransferSetting' (from MAPI) or 'FileSystem' (from arclftctl --xml)
        self.fileSystemName = None
        self.largeFileTransfer = None
        self.lowerLimitInMegaBytes = None
        self.transferPartMinInMegaBytes = None
        self.xml = None

        if lftelm is None:
            return

        if lftelm.tag != 'LargeFileTransferSetting' and lftelm.tag != 'FileSystem':
            log_error("A wrong XML element was passed for LargeFileTransferInfo. tag=%s" % lftelm.tag)
            raise ParameterError("A wrong XML element for LargeFileTransferInfo. tag=%s" % lftelm.tag)

        if lftelm.tag == 'LargeFileTransferSetting':
            e = lftelm.find('largeFileTransfer')
            self.largeFileTransfer = e.text if e is not None else None
            e = lftelm.find('lowerLimitInMegaBytes')
            self.lowerLimitInMegaBytes = (int(e.text) if e.text is not None else None) if e is not None else None
        else:
            try:
                self.fileSystemName = lftelm.attrib['fileSystemName']
                self.largeFileTransfer = lftelm.attrib['largeFileTransfer']
                value = lftelm.attrib['lowerLimitInMegaBytes']
                self.lowerLimitInMegaBytes = int(value) if len(value) != 0 else None
                value = lftelm.attrib['transferPartMinInMegaBytes']
                self.transferPartMinInMegaBytes = int(value) if len(value) != 0 else None
            except KeyError, e:
                log_error("Expected information is missing in arclftctl XML output. name=%s" % str(e))
                raise FormatError("Expected information is missing in arclftctl XML output. name=%s" % str(e))

        self.xml = ET.tostring(lftelm, encoding='UTF-8')

        # largeFileTransfer tag can be empty when AFM is disabled (maybe bug of MAPI...)
        if self.largeFileTransfer not in ('Enable', 'Disable'):
            # Assume it's disabled and has default value
            self.largeFileTransfer = 'Disable'
            self.lowerLimitInMegaBytes = 100

    def has_different_configuration_from(self, lftinfo):
        # Usage
        if lftinfo.largeFileTransfer != self.largeFileTransfer:
            log_info("Large File Transfer usage has been changed. previous:%s, current:%s" % (lftinfo.largeFileTransfer, self.largeFileTransfer))
            return True

        # Lower limit
        if lftinfo.lowerLimitInMegaBytes != self.lowerLimitInMegaBytes:
            log_info("Lower limit of file size for Large File Transfer has been changed. previous:%s, current:%s" % (lftinfo.lowerLimitInMegaBytes, self.lowerLimitInMegaBytes))
            return True

        # Part size
        if lftinfo.transferPartMinInMegaBytes != self.transferPartMinInMegaBytes:
            log_info("Minimum transfer part size for Large File Transfer has been changed. previous:%s, current:%s" % (lftinfo.transferPartMinInMegaBytes, self.transferPartMinInMegaBytes))
            return True

        return False

class FileSystemInfo(object):
    def __init__(self, fselm=None):
        # fselm is an XML element of 'FileSystem'
        self.fileSystemName = ""
        self.usage = []
        self.deviceStatus = None
        self.freeBlocksInMegaBytes = None
        self.usedBlocksInMegaBytes = None
        self.totalBlocksInMegaBytes = None
        self.isLvmEnabled = None
        self.volumeGroupName = None
        self.fileSystemSizeInGigaByte = None
        self.workSpaceSizeInGigaByte = None
        self.isLvmStripingEnabled = None
        self.stripes = None
        self.stripeSize = None
        self.aclType = None
        self.NamespaceShareSettings = None
        self.UseVersioning = None
        self.periodToHold = None
        self.customSchedule = None
        self.customSchedule15Minute = None
        self.customScheduleHourly = None
        self.customScheduleDaily = None
        self.customScheduleWeekly = None
        self.customScheduleMonthly = None
        self.customScheduleYearly = None
        self.isBypassEnabled = None
        self.WormSetting = None
        self.MountSetting = None
        self.LUs = None
        self.WorkSpaceLUs = None
        self.LargeFileTransferSetting = None
        self.xml = None

        if fselm is None:
            return

        if fselm.tag != 'FileSystem':
            log_error("A wrong XML element was passed for FileSystemInfo. tag=%s" % fselm.tag)
            raise ParameterError("A wrong XML element for FileSystemInfo. tag=%s" % fselm.tag)

        e = fselm.find('fileSystemName')
        self.fileSystemName = e.text if e is not None else None
        e = fselm.find('usage')
        self.usage = map(str.strip, e.text.split(',')) if e is not None else None
        e = fselm.find('deviceStatus')
        self.deviceStatus = e.text if e is not None else None
        e = fselm.find('freeBlocksInMegaBytes')
        self.freeBlocksInMegaBytes = (float(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('usedBlocksInMegaBytes')
        self.usedBlocksInMegaBytes = (float(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('totalBlocksInMegaBytes')
        self.totalBlocksInMegaBytes = (float(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('isLvmEnabled')
        self.isLvmEnabled = (Utility.str2bool(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('volumeGroupName')
        self.volumeGroupName = e.text if e is not None else None
        e = fselm.find('fileSystemSizeInGigaByte')
        # Convert to ingeter
        self.fileSystemSizeInGigaByte = (int(float(e.text)) if e.text is not None else None) if e is not None else None
        e = fselm.find('workSpaceSizeInGigaByte')
        # Convert to ingeter
        self.workSpaceSizeInGigaByte = (int(float(e.text)) if e.text is not None else None) if e is not None else None
        e = fselm.find('isLvmStripingEnabled')
        self.isLvmStripingEnabled = (Utility.str2bool(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('stripes')
        self.stripes = (int(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('stripeSize')
        self.stripeSize = (int(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('aclType')
        self.aclType = e.text if e is not None else None
        e = fselm.find('NamespaceShareSettings')
        self.NamespaceShareSettings = NamespaceInfo(e) if e is not None else None
        e = fselm.find('UseVersioning')
        self.UseVersioning = e.text if e is not None else None
        e = fselm.find('periodToHold')
        self.periodToHold = (int(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('isBypassEnabled')
        self.isBypassEnabled = (Utility.str2bool(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('customSchedule')
        self.customSchedule = e.text if e is not None else None
        e = fselm.find('customSchedule15Minute')
        # 0 is set if this is not used
        self.customSchedule15Minute = (int(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('customScheduleHourly')
        # 0 is set if this is not used
        self.customScheduleHourly = (int(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('customScheduleDaily')
        # 0 is set if this is not used
        self.customScheduleDaily = (int(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('customScheduleWeekly')
        # 0 is set if this is not used
        self.customScheduleWeekly = (int(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('customScheduleMonthly')
        # 0 is set if this is not used
        self.customScheduleMonthly = (int(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('customScheduleYearly')
        # 0 is set if this is not used
        self.customScheduleYearly = (int(e.text) if e.text is not None else None) if e is not None else None
        e = fselm.find('WormSetting')
        self.WormSetting = WormInfo(e) if e is not None else None
        e = fselm.find('MountSetting')
        self.MountSetting = MountInfo(e) if e is not None else None
        el = fselm.find('LUs')
        if el is not None:
            self.LUs = []
            for e in el.findall('LU'):
                self.LUs.append(FileSystemLUInfo(e))
        el = fselm.find('WorkSpaceLUs')
        if el is not None:
            self.WorkSpaceLUs = []
            for e in el.findall('WorkSpace'):
                self.WorkSpaceLUs.append(WorkSpaceLUInfo(e))
        e = fselm.find('LargeFileTransferSetting')
        self.LargeFileTransferSetting = LargeFileTransferInfo(e) if e is not None else None

        self.xml = ET.tostring(fselm, encoding='UTF-8')

        if self._sanity_check():
            log_error("XML data (reconstructed): %s" % ET.tostring(fselm))
            raise FormatError("Expected information is missing in FileSystem XML.")

    def _sanity_check(self):
        error = False
        if self.fileSystemName is None:
            log_error("File system name is not defined. fileSystemName is empty or missing.")
            error = True
        if self.usage is None:
            log_error("Usage is not defined. usage is empty or missing.")
            error = True
        if self.fileSystemSizeInGigaByte is None:
            log_error("File system size is not defined. fileSystemSizeInGigaByte is empty or missing.")
            error = True
        if self.aclType is None:
            log_error("ACL type is not defined. aclType is empty or missing.")
            error = True
        if self.NamespaceShareSettings is None:
            log_error("Namespace is not defined. NamespaceShareSettings is empty or missing.")
            error = True
        return error

    def toxml(self, model=None, version=None, fslulist=None, wslulist=None, fs_capacity=None, ws_capacity=None, is_likeforlike=False):
        log_info("Building an XML for file system.")

        # Check if this is a supported file system
        if not self.is_supported():
            log_error("Tried to build an XML for a file system that is not supported.")
            raise UnsupportedError("Tried to build an XML for a file system that is not supported.")

        ### Common ###
        fs = ET.Element("FileSystem")

        # File system name
        ET.SubElement(fs, "fileSystemName").text = self.fileSystemName

        # ACL type
        ET.SubElement(fs, "aclType").text = self.aclType

        # Versioning
        if self.UseVersioning is None:
            log_warning("No versioning is defined. 'used' will be set.")
            value = "used"
        else:
            value = self.UseVersioning
        ET.SubElement(fs, "UseVersioning").text = value

        # Period to hold
        if self.periodToHold is None:
            log_warning("Period to hold is not defined. '36500' will be set.")
            value = 36500
        else:
            value = self.periodToHold
        ET.SubElement(fs, "periodToHold").text = str(value)

        # CIFS bypass traverse cheking
        if self.isBypassEnabled is None:
            log_warning("CIFS bypass traverse checking is not defined. 'true' will be set.")
            value = "true"
        else:
            value = Utility.bool2str(self.isBypassEnabled)
        ET.SubElement(fs, "isBypassEnabled").text = value

        # Custom Schedule
        if self.customSchedule is None:
            log_warning("Usage of Custom Schedule is not defined. 'not used' will be set.")
            value = "not used"
        elif sum([x is not None for x in [self.customSchedule15Minute, self.customScheduleHourly, self.customScheduleDaily, self.customScheduleWeekly, self.customScheduleMonthly, self.customScheduleYearly]]) == 0:
            # All of customSchedulexxxly is empty or not defined? Disable Custom Schedule for safety.
            log_warning("No schedule is defined. 'not used' will be set for Custom Schedule.")
            value = "not used"
        else:
            value = self.customSchedule
        ET.SubElement(fs, "customSchedule").text = value
        if self.customSchedule == "used":
            ET.SubElement(fs, "customSchedule15Minute").text = str(self.customSchedule15Minute)
            ET.SubElement(fs, "customScheduleHourly").text = str(self.customScheduleHourly)
            ET.SubElement(fs, "customScheduleDaily").text = str(self.customScheduleDaily)
            ET.SubElement(fs, "customScheduleWeekly").text = str(self.customScheduleWeekly)
            ET.SubElement(fs, "customScheduleMonthly").text = str(self.customScheduleMonthly)
            ET.SubElement(fs, "customScheduleYearly").text = str(self.customScheduleYearly)

        # Namespace settings
        ns = ET.SubElement(fs, "NamespaceShareSettings")
        # Namespace type
        ET.SubElement(ns, "namespaceType").text = self.NamespaceShareSettings.namespaceType
        # type
        ET.SubElement(ns, "type").text = self.NamespaceShareSettings.type

        # Other parameters of NamespaceShareSettings are not required for a regular file system

        # WORM
        if self.WormSetting.maxRetention is not None:
            # WORM is enabled if maxRetention is defined
            worm = ET.SubElement(fs, "WormSetting")
            ET.SubElement(worm, "maxRetention").text = self.WormSetting.maxRetention
            ET.SubElement(worm, "minRetention").text = self.WormSetting.minRetention
            ET.SubElement(worm, "isAutoCommitEnabled").text = Utility.bool2str(self.WormSetting.isAutoCommitEnabled)
            if self.WormSetting.isAutoCommitEnabled:
                ET.SubElement(worm, "commitModeSetting").text = self.WormSetting.commitModeSetting
                ET.SubElement(worm, "autoCommitPeriod").text = self.WormSetting.autoCommitPeriod
                ET.SubElement(worm, "defaultRetention").text = self.WormSetting.defaultRetention

        # No LFT - LFT will be configured separately

        ### Model specific parameters ###
        if model == Model.CLUSTER:
            # For cluster

            # LVM
            if self.isLvmEnabled is None:
                # Primary node may be a single/VM node
                log_info("LVM is not defined. 'true' will be set.")
                value = "true"
            else:
                value = Utility.bool2str(self.isLvmEnabled)
            ET.SubElement(fs, "isLvmEnabled").text = value

            # LVM striping
            # Like for Like mode only. Otherwise LU capacity will not meet the requirement.
            if is_likeforlike:
                if self.isLvmStripingEnabled is None:
                    # Primary node may be a single/VM node
                    log_info("LVM striping is not defined. 'false' will be set.")
                    value = "false"
                else:
                    value = Utility.bool2str(self.isLvmStripingEnabled)
                ET.SubElement(fs, "isLvmStripingEnabled").text = value
                if self.isLvmStripingEnabled:
                    if self.stripeSize is None:
                        log_warning("Stripe size is not defined while striping is enabled. '64' will be set.")
                        value = 64
                    else:
                        value = self.stripeSize
                    ET.SubElement(fs, "stripeSize").text = str(self.stripeSize)
            else:
                # LVM striping is disabled by default
                pass

            # LUs
            lus = ET.SubElement(fs, "LUs")
            if fslulist is None:
                log_error("No LUs are specified for cluster configuration.")
                raise ParameterError("No LUs are specified for building an XML for cluster configuration.")
            for luname in [lu.deviceFileName for lu in fslulist]:
                lu = ET.SubElement(lus, "LU")
                ET.SubElement(lu, "deviceFileName").text = luname

            # Work space LUs
            if "ActMig" in self.usage:
                wslus = ET.SubElement(fs, "WorkSpaceLUs")
                # wslulist can be None in case AFM is not used
                for luname in [lu.deviceFileName for lu in wslulist]:
                    ws = ET.SubElement(wslus, "WorkSpace")
                    ET.SubElement(ws, "deviceFileName").text = luname
        elif model == Model.SINGLE:
            # For single

            # Volume group
            if is_likeforlike:
                # Use the same volume group
                if self.volumeGroupName is None:
                    log_error("Volume group name is not defined. It must be defined for Like for Like mode.")
                    raise FormatError("Volume group name is not defined for Like for Like mode.")
                ET.SubElement(fs, "volumeGroupName").text = self.volumeGroupName
            else:
                # Let HDI choose one
                pass

            # File system capacity
            value = fs_capacity
            ET.SubElement(fs, "fileSystemSizeInGigaByte").text = str(value)

            # Work space capacity
            if "ActMig" in self.usage:
                value = ws_capacity
                ET.SubElement(fs, "workSpaceSizeInGigaByte").text = str(value)

        # Does not add standalone="yes" as there is no easy way to add it with ElementTree..
        return ET.tostring(fs, encoding='UTF-8')

    def is_supported(self):
        """ Check if the file system is supported for failover.

        Not supported file systems are:
            Read-only Content Sharing
            Read-Write Content Sharing
            Roaming Home Directory
            Subtree file system
            Not mounted file system

        Args:
            None
        Returns:
            bool: True if the file system is supported. False if not.
        """
        # Is this a content sharing file system?
        if self.NamespaceShareSettings.type != 'Read/Write':
            # Content Sharing file system is not supported.
            log_info("File system %s is not supported because it is a Content Sharing file system or not linked to HCP. (type = %s)" % (self.fileSystemName, self.NamespaceShareSettings.type))
            return False

        # Is this a subtree file system?
        if self.NamespaceShareSettings.namespaceType != 'FileSystem':
            # Subtree file system is not supported because Subtree file system
            # cannot be restored on a different HDI system.
            log_info("File system %s is not supported because it is Subtree file system. (namespaceType = %s)" % (self.fileSystemName, self.NamespaceShareSettings.namespaceType))
            return False

        # Is this mounted?
        if not self.MountSetting.isMounted:
            # Unmounted file system is out of scope because some information is not defined when the file system is not mounted
            log_info("File system %s is not supported because it is not mounted. (isMounted = %s)" % (self.fileSystemName, self.MountSetting.isMounted))
            return False

        return True

    def has_different_configuration_from(self, fsinfo):
        """ Compares with a FileSystemInfo and checks if there are any differences.

        This function compares each 'changable' parameter between myself and
        another FileSystemInfo instance provided as 'fsinfo'.

        NOTE: This function does NOT check difference of Large File Transfer
        settings because this function is intented to be used to decide whether
        a file system needs to be synchronized again. LFT settings are not set
        during synchronization but after data restoration (i.e. arcrestore).

        Args:
            fsinfo (FileSystemInfo): FileSystemInfo instance to be compared with.
        Returns:
            bool: True if there is any difference. Otherwise False.
        """

        # usage
        prev_usage_set = set(fsinfo.usage)
        curr_usage_set = set(self.usage)
        if prev_usage_set != curr_usage_set:
            log_info("Usage has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.usage, self.usage))
            return True

        # LVM
        if fsinfo.isLvmEnabled != self.isLvmEnabled:
            log_info("LVM usage has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.isLvmEnabled, self.isLvmEnabled))
            return True

        # File system size
        if fsinfo.fileSystemSizeInGigaByte != self.fileSystemSizeInGigaByte:
            log_info("File system capacity has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.fileSystemSizeInGigaByte, self.fileSystemSizeInGigaByte))
            return True

        # Work space size
        if fsinfo.workSpaceSizeInGigaByte != self.workSpaceSizeInGigaByte:
            log_info("Work space capacity has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.workSpaceSizeInGigaByte, self.workSpaceSizeInGigaByte))
            return True

        # LVM stripe usage
        if fsinfo.isLvmStripingEnabled != self.isLvmStripingEnabled:
            log_info("LVM striping usage has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.isLvmStripingEnabled, self.isLvmStripingEnabled))
            return True

        # ACL type
        if fsinfo.aclType != self.aclType:
            log_info("ACL type has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.aclType, self.aclType))
            return True

        # Does not check NamespaceShareSettings.
        # Only one setting of NamespaceShareSettings is supported.

        # Versioning usage
        if fsinfo.UseVersioning != self.UseVersioning:
            log_info("Versioning usage has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.UseVersioning, self.UseVersioning))
            return True

        # Period to hold
        if fsinfo.periodToHold != self.periodToHold:
            log_info("Period to hold has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.periodToHold, self.periodToHold))
            return True

        # CIFS bypass traverse checking
        if fsinfo.isBypassEnabled != self.isBypassEnabled:
            log_info("CIFS bypass traverse checking usage has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.isBypassEnabled, self.isBypassEnabled))
            return True

        # Custom Schedule usage
        if fsinfo.customSchedule != self.customSchedule:
            log_info("Custom Schedule usage has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.customSchedule, self.customSchedule))
            return True

        # Custom Schedule schedule
        if fsinfo.customSchedule15Minute != self.customSchedule15Minute:
            log_info("Custom Schedule 15-Minute schedule has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.customSchedule15Minute, self.customSchedule15Minute))
            return True
        if fsinfo.customScheduleHourly != self.customScheduleHourly:
            log_info("Custom Schedule hourly schedule has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.customScheduleHourly, self.customScheduleHourly))
            return True
        if fsinfo.customScheduleDaily != self.customScheduleDaily:
            log_info("Custom Schedule daily schedule has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.customScheduleDaily, self.customScheduleDaily))
            return True
        if fsinfo.customScheduleWeekly != self.customScheduleWeekly:
            log_info("Custom Schedule weekly schedule has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.customScheduleWeekly, self.customScheduleWeekly))
            return True
        if fsinfo.customScheduleMonthly != self.customScheduleMonthly:
            log_info("Custom Schedule monthly schedule has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.customScheduleMonthly, self.customScheduleMonthly))
            return True
        if fsinfo.customScheduleYearly != self.customScheduleYearly:
            log_info("Custom Schedule yearly schedule has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.customScheduleYearly, self.customScheduleYearly))
            return True

        # WORM
        if self.WormSetting.has_different_configuration_from(fsinfo.WormSetting):
            log_info("Worm setting has been changed for %s." % self.fileSystemName)
            return True

        # Mount setting
        if fsinfo.MountSetting.isFileLastAccessTimeRecordingEnabled != self.MountSetting.isFileLastAccessTimeRecordingEnabled:
            log_info("Last file system access time recording setting has been changed for %s. previous:%s, current:%s" % (self.fileSystemName, fsinfo.MountSetting.isFileLastAccessTimeRecordingEnabled, self.MountSetting.isFileLastAccessTimeRecordingEnabled))
            return True

        return False

    @classmethod
    def calculate_workspace_capacity(cls, model, fs_capacity, lft=False):
        if lft:
            if model == Model.CLUSTER:
                ws_capacity = cls._calc_ws_lft_cluster(fs_capacity)
            elif model == Model.SINGLE:
                ws_capacity = cls._calc_ws_lft_single(fs_capacity)
        else:
            ws_capacity = cls._calc_ws_nolft(fs_capacity)

        return ws_capacity

    @classmethod
    def _calc_ws_nolft(cls, fs_capacity):
        # fs_capacity, ws_capacity are GB unit
        # From Installation and Configuration Guide
        if fs_capacity <= 17 * 1024:
            ws_capacity = 10
        elif fs_capacity > 17 * 1024 and fs_capacity <= 256 * 1024:
            ws_capacity = 25
        elif fs_capacity > 256 * 1024:
            ws_capacity = 50
        return ws_capacity

    @classmethod
    def _calc_ws_lft_cluster(cls, fs_capacity):
        # fs_capacity, ws_capacity are GB unit
        # From LFT SRT material
        if fs_capacity < 4 * 1024:
            ws_capacity = 300
        elif fs_capacity >= 4 * 1024 and fs_capacity < 64 * 1024:
            ws_capacity = 350
        elif fs_capacity >= 64 * 1024 and fs_capacity < 256 * 1024:
            ws_capacity = 1130
        elif fs_capacity >= 256 * 1024 and fs_capacity < 512 * 1024:
            ws_capacity = 2155
        elif fs_capacity >= 512 * 1024 and fs_capacity < 1024 * 1024:
            ws_capacity = 5230
        return ws_capacity

    @classmethod
    def _calc_ws_lft_single(cls, fs_capacity):
        # fs_capacity, ws_capacity are GB unit
        # From LFT SRT material
        if fs_capacity < 1 * 1024:
            ws_capacity = 40
        elif fs_capacity >= 1 * 1024 and fs_capacity < 25 * 1024:
            ws_capacity = 50
        elif fs_capacity >= 25 * 1024 and fs_capacity < 40 * 1024:
            ws_capacity = 60
        elif fs_capacity >= 40 * 1024:
            ws_capacity = 80
        return ws_capacity

class CIFSShareInfo(object):
    def __init__(self, cifselm=None):
        # cifselm is an XML element of 'CIFSShare'
        # Accepts both MAPI XML and cifsbackup XML
        # Contains only a part of information
        self.exportPoint = None
        self.cifsShareName = None
        self.filesystem = None
        self.xml = None

        if cifselm is None:
            return

        if cifselm.tag != 'CIFSShare':
            log_error("A wrong XML element was passed for CIFSShareInfo. tag=%s" % cifselm.tag)
            raise ParameterError("A wrong XML element for CIFSShareInfo. tag=%s" % cifselm.tag)

        try:
            if cifselm.find('exportPoint') is None:
                self.exportPoint = cifselm.attrib['exportPoint']
                self.cifsShareName = cifselm.attrib['cifsShareName']
            else:
                self.exportPoint = cifselm.find('exportPoint').text
                self.cifsShareName = cifselm.find('cifsShareName').text
        except (KeyError, AttributeError), e:
            log_error("Expected information is missing in XML for a CIFS share. error=%s" % str(e))
            raise FormatError("Expected information is missing in XML for a CIFS share.")
        m = re.match(r'^/mnt/(\w+)(/.*)?', self.exportPoint)
        if m is None:
            log_error("Couldn't find file system name from export point. exportPoint=%s" % self.exportPoint)
            raise FormatError("A wrong export point format in a CIFS share XML.")
        self.filesystem = m.group(1)

        self.xml = ET.tostring(cifselm, encoding='UTF-8')

class NFSShareInfo(object):
    def __init__(self, nfselm=None):
        # nfsselm is an XML element of 'NFSShare'
        # Accepts both MAPI XML and nfsbackup XML
        # Contains only a part of information
        self.exportPoint = None
        self.filesystem = None
        self.xml = None

        if nfselm is None:
            return

        if nfselm.tag != 'NFSShare':
            log_error("A wrong XML element was passed for NFSShareInfo. tag=%s" % nfselm.tag)
            raise ParameterError("A wrong XML element for NFSShareInfo. tag=%s" % nfselm.tag)

        try:
            if nfselm.find('exportPoint') is None:
                self.exportPoint = nfselm.attrib['exportPoint']
            else:
                self.exportPoint = nfselm.find('exportPoint').text
        except (KeyError, AttributeError), e:
            log_error("Expected information is missing in XML for a NFS share. error=%s" % str(e))
            raise FormatError("Expected information is missing in XML for a NFS share.")
        m = re.match(r'^/mnt/(\w+)(/.*)?', self.exportPoint)
        if m is None:
            log_error("Couldn't find file system name from export point. exportPoint=%s" % self.exportPoint)
            raise FormatError("A wrong export point format in a NFS share XML.")
        self.filesystem = m.group(1)

        self.xml = ET.tostring(nfselm, encoding='UTF-8')

class SubtreeNamespaceInfo(object):
    def __init__(self, stelm=None):
        # stelm is an XML element of 'namespace'
        self.name = None
        self.directory = None
        self.namespaceQuota = None
        self.xml = None

        if stelm is None:
            return

        if stelm.tag != 'namespace':
            log_error("A wrong XML element was passed for SubtreeNamespaceInfo. tag=%s" % stelm.tag)
            raise ParameterError("A wrong XML element for SubtreeNamespaceInfo. tag=%s" % stelm.tag)

        e = stelm.find('name')
        self.name = e.text if e is not None else None
        e = stelm.find('directory')
        self.directory = e.text if e is not None else None
        e = stelm.find('namespaceQuota')
        self.namespaceQuota = e.text if e is not None else None

        self.xml = ET.tostring(stelm, encoding='UTF-8')

class MigTaskInfo(object):
    def __init__(self, taskelm=None):
        # taskelm is an XML element of 'MigrationTask'
        self.task_name = None
        self.task_priority = None
        self.task_comment = None
        self.task_enabled = None
        self.task_status = None
        self.fileSystemName = None
        self.namespace = None
        self.Namespaces = None
        self.schedule_start_time = None
        self.schedule_interval = None
        self.duration = None
        self.xml = None

        if taskelm is None:
            return

        if taskelm.tag != 'MigrationTask':
            log_error("A wrong XML element was passed for MigrationTask. tag=%s" % taskelm.tag)
            raise ParameterError("A wrong XML element for MigrationTask. tag=%s" % taskelm.tag)

        e = taskelm.find('task-name')
        self.task_name = e.text if e is not None else None
        e = taskelm.find('task-priority')
        self.task_priority = int(e.text) if e.text is not None else None
        e = taskelm.find('task-comment')
        self.task_comment = e.text if e is not None else None
        e = taskelm.find('task-enabled')
        self.task_enabled = e.text if e is not None else None
        e = taskelm.find('task-status')
        self.task_status = e.text if e is not None else None
        e = taskelm.find('fileSystemName')
        self.fileSystemName = e.text if e is not None else None
        e = taskelm.find('namespace')
        self.namespace = e.text if e is not None else None
        el = taskelm.find('Namespaces')
        if el is not None:
            self.Namespaces = []
            for e in el.findall('namespace'):
                self.Namespaces.append(SubtreeNamespaceInfo(e))
        e = taskelm.find('schedule-start-time')
        # Store as epoch time for easier handling
        self.schedule_start_time = Utility.iso8601_to_epoch(e.text) if e is not None else None
        e = taskelm.find('schedule-interval')
        self.schedule_interval = e.text if e is not None else None
        e = taskelm.find('duration')
        self.duration = int(e.text) if e.text is not None else None

        if self._sanity_check():
            log_error("XML data (reconstructed): %s" % ET.tostring(taskelm))
            raise FormatError("Expected information is missing in MigrationTask XML.")

        self.xml = ET.tostring(taskelm, encoding='UTF-8')

    def toxml(self, start_time):
        # start_time is used as schedule-start-time
        # start_time has to be presented as epoch time (integer)
        log_info("Building an XML for migration task.")

        task = ET.Element("MigrationTask")

        # Task name
        ET.SubElement(task, "task-name").text = self.task_name

        # Task priority
        if self.task_priority is None:
            log_warning("Task priority is not defined. 'middle' will be set.")
            value = 'middle'
        else:
            value = self.task_priority
        ET.SubElement(task, "task-priority").text = str(value)

        # Task comment
        if self.task_comment is not None:
            ET.SubElement(task, "task-comment").text = self.task_comment

        # File system name
        ET.SubElement(task, "fileSystemName").text = self.fileSystemName

        # Namespace
        if self.namespace is not None:
            # "FileSystem"
            ET.SubElement(task, "namespace").text = self.namespace
        else:
            # "Subtree"
            # Not implemented yet!
            log_error("Building migration task XML with Subtree namespaces is not implemented.")
            raise UnsupportedError("Subtree namespaces is not supported.")

        # Schedule start time
        ET.SubElement(task, "schedule-start-time").text = Utility.epoch_to_iso8601(start_time)

        # Schedule interval
        vals = self.schedule_interval.split(' ')
        num = vals[0]
        orig_unit = vals[1]
        if orig_unit == 'minutes':
            unit = 'm'
        elif orig_unit in ('hour', 'hours'):
            unit = 'h'
        elif orig_unit in ('day', 'days'):
            unit = 'd'
        elif orig_unit in ('week', 'weeks'):
            unit = 'w'
        elif orig_unit in ('month', 'months'):
            unit = 'M'
        ET.SubElement(task, "schedule-interval").text = "%s%s" % (num, unit)

        # Duration
        if self.duration is None:
            log_warning("Duration is not defined. 0 will be set.")
            value = 0
        else:
            value = self.duration
        ET.SubElement(task, "duration").text = str(value)

        return ET.tostring(task, encoding='UTF-8')

    def _sanity_check(self):
        log_info("Sanity checking of migration task XML.")
        error = False
        if self.task_name is None:
            log_error("Task name is not defined. task-name is empty or missing.")
            error = True
        if self.fileSystemName is None:
            log_error("File system name is not defined. fileSystemName is empty or missing.")
            error = True
        if self.namespace is None and (self.Namespaces is None or len(self.Namespaces) == 0):
            log_error("Namespace information is not defined. namspace or Namespaces is empty of missing.")
            error = True
        if self.schedule_interval is None:
            log_error("Schedule interval is not defined. schedule-interval is empty of missing.")
            error = True
        return error

class PhysicalNodeInfo(object):
    def __init__(self, nelm=None):
        self.location = None
        self.hostName = None
        self.physicalNodeName = None
        self.status = None
        self.managementLANAddress = None
        self.managementLANAddressIPv6 = None
        self.HBLanAddress = None
        self.BMCLanAddress = None
        self.resourceGroupStatus = None
        self.resourceGroupErrorInfo = None
        self.runningNode = None
        self.serial = None
        self.SystemVesion = None    # Dictionary
        self.xml = None

        if nelm is None:
            return

        if nelm.tag != 'PhysicalNode':
            log_error("A wrong XML element was passed for PhysicalNode. tag=%s" % nelm.tag)
            raise ParameterError("A wrong XML element for PhysicalNode. tag=%s" % nelm.tag)

        e = nelm.find('location')
        self.location = e.text if e is not None else None
        self.hostName = nelm.find('hostName').text
        e = nelm.find('physicalNodeName')
        self.physicalNodeName = e.text if e is not None else None
        e = nelm.find('status')
        self.status = e.text if e is not None else None
        e = nelm.find('managementLANAddress')
        self.managementLANAddress = e.text if e is not None else None
        e = nelm.find('managementLANAddressIPv6')
        self.managementLANAddressIPv6 = e.text if e is not None else None
        e = nelm.find('HBLanAddress')
        self.HBLanAddress = e.text if e is not None else None
        e = nelm.find('BMCLanAddress')
        self.BMCLanAddress = e.text if e is not None else None
        e = nelm.find('resourceGroupStatus')
        self.resourceGroupStatus = e.text if e is not None else None
        e = nelm.find('resourceGroupErrorInfo')
        self.resourceGroupErrorInfo = e.text if e is not None else None
        e = nelm.find('runningNode')
        self.runningNode = e.text if e is not None else None
        e = nelm.find('serial')
        self.serial = e.text if e is not None else None
        sysver = nelm.find('SystemVersion')
        self.SystemVersion = {'name': sysver.find('name').text, 'version': sysver.find('version').text} if sysver is not None else None

        self.xml = ET.tostring(nelm, encoding='UTF-8')

class ProcessingNodeInfo(object):
    def __init__(self, nodeelm=None):
        # nodeelm is an XML element of 'ProcessingNode'
        self.SystemConfiguration = None # Dictionary
        self.status = None
        self.name = None
        self.LU = None  # Dictionary
        self.PhysicalNodeSelf = None
        self.PhysicalNodeOther = None
        self.xml = None

        if nodeelm is None:
            return

        if nodeelm.tag != 'ProcessingNode':
            log_error("A wrong XML element was passed for ProcessingNode. tag=%s" % nodeelm.tag)
            raise ParameterError("A wrong XML element for ProcessingNode. tag=%s" % nodeelm.tag)

        sysconf = nodeelm.find('SystemConfiguration')
        self.SystemConfiguration = {'mode': sysconf.find('mode').text} if sysconf is not None else None
        e = nodeelm.find('status')  # Cluster status
        self.status = e.text if e is not None else None
        e = nodeelm.find('name')    # Cluster name
        self.name = e.text if e is not None else None
        lu = nodeelm.find('LU')     # Cluster management LU
        self.LU = {'model': lu.find('model').text, 'serial': lu.find('serial').text, 'ldev': lu.find('ldev').text} if lu is not None else None
        nodes = nodeelm.findall('PhysicalNode')
        # Assuming self node is output at first
        self.PhysicalNodeSelf = PhysicalNodeInfo(nodes[0]) if len(nodes) > 0 else None
        self.PhysicalNodeOther = PhysicalNodeInfo(nodes[1]) if len(nodes) == 2 else None

        self.xml = ET.tostring(nodeelm, encoding='UTF-8')

class LUInfo(object):
    def __init__(self, luelm=None):
        # luelm is an XML element of 'LU' of LU resource
        # Get only necessary information
        self.deviceFileName = None
        self.size = None
        self.function = None
        self.usedBy = None
        self.xml = None

        if luelm is None:
            return

        if luelm.tag != 'LU':
            log_error("A wrong XML element was passed for LU. tag=%s" % luelm.tag)
            raise ParameterError("A wrong XML element for LU. tag=%s" % luelm.tag)

        e = luelm.find('deviceFileName')
        self.deviceFileName = e.text if e is not None else None
        e = luelm.find('size')
        self.size = (int(float(e.text)) if e.text != "-" else None) if e is not None else None
        e = luelm.find('function')
        self.function = e.text if e is not None else None
        e = luelm.find('usedBy')
        self.usedBy = e.text if e is not None else None

        self.xml = ET.tostring(luelm, encoding='UTF-8')

class VolumeGroupInfo(object):
    def __init__(self, vgelm=None):
        # vgelm is an XML element of 'VolumeGroup'
        self.volumeGroupName = None
        self.totalSizeInGigaByte = None
        self.freeSizeInGigaByte = None
        self.deviceStatus = None
        self.xml = None

        if vgelm is None:
            return

        if vgelm.tag != 'VolumeGroup':
            log_error("A wrong XML element was passed for VolumeGroup. tag=%s" % vgelm.tag)
            raise ParameterError("A wrong XML element for VolumeGroup. tag=%s" % vgelm.tag)

        e = vgelm.find('volumeGroupName')
        self.volumeGroupName = e.text if e is not None else None
        e = vgelm.find('totalSizeInGigaByte')
        self.totalSizeInGigaByte = int(e.text) if e is not None else None
        e = vgelm.find('freeSizeInGigaByte')
        self.freeSizeInGigaByte = int(e.text) if e is not None else None
        e = vgelm.find('deviceStatus')
        self.deviceStatus = e.text if e is not None else None

        self.xml = ET.tostring(vgelm, encoding='UTF-8')

# Functions

def parse_filesystems(xml_data=None, xml_file=None):
    return _parse_xml("FileSystems", xml_data=xml_data, xml_file=xml_file)

def parse_filesystem(xml_data=None, xml_file=None):
    return _parse_xml("FileSystem", xml_data=xml_data, xml_file=xml_file)

def parse_mountsetting(xml_data=None, xml_file=None):
    return _parse_xml("MountSetting", xml_data=xml_data, xml_file=xml_file)

def parse_arclftctl(xml_data=None, xml_file=None):
    return _parse_xml("LargeFileTransfer", xml_data=xml_data, xml_file=xml_file)

def parse_arclftctl_one(xml_data=None, xml_file=None):
    return _parse_xml("LargeFileTransferOne", xml_data=xml_data, xml_file=xml_file)

def parse_cifsshares(xml_data=None, xml_file=None):
    return _parse_xml("CIFSShares", xml_data=xml_data, xml_file=xml_file)

def parse_cifsshare(xml_data=None, xml_file=None):
    return _parse_xml("CIFSShare", xml_data=xml_data, xml_file=xml_file)

def parse_nfsshares(xml_data=None, xml_file=None):
    return _parse_xml("NFSShares", xml_data=xml_data, xml_file=xml_file)

def parse_nfsshare(xml_data=None, xml_file=None):
    return _parse_xml("NFSShare", xml_data=xml_data, xml_file=xml_file)

def parse_migtasks(xml_data=None, xml_file=None):
    return _parse_xml("MigrationTasks", xml_data=xml_data, xml_file=xml_file)

def parse_migtask(xml_data=None, xml_file=None):
    return _parse_xml("MigrationTask", xml_data=xml_data, xml_file=xml_file)

def parse_interfaces(xml_data=None, xml_file=None):
    return _parse_xml("NetworkInterfaces", xml_data=xml_data, xml_file=xml_file)

def parse_resident(xml_data=None, xml_file=None):
    return _parse_xml("CacheResident", xml_data=xml_data, xml_file=xml_file)

def parse_processing_node(xml_data=None, xml_file=None):
    return _parse_xml("ProcessingNode", xml_data=xml_data, xml_file=xml_file)

def parse_lus(xml_data=None, xml_file=None):
    return _parse_xml("LUs", xml_data=xml_data, xml_file=xml_file)

def parse_vg(xml_data=None, xml_file=None):
    return _parse_xml("VolumeGroups", xml_data=xml_data, xml_file=xml_file)

def parse_hostname(txt_file):
    with open(txt_file, 'r') as f:
        return f.read().strip()

def parse_uuid(txt_file):
    with open(txt_file, 'r') as f:
        return f.read().strip()

def _parse_xml(resource, xml_data=None, xml_file=None):
    # Read XML string or file and return a list of the resource objects

    # Accepts only either xml_data or xml_file
    if xml_data is None and xml_file is None:
        log_error("No XML source was specified for _parse_xml().")
        raise ParameterError("No XML source was specified for _parse_xml().")
    if xml_data is not None and xml_file is not None:
        log_error("Too many XML sources were specified for _parse_xml().")
        raise ParameterError("Too many XML sources were specified for _parse_xml().")

    # Parse XML data
    try:
        if xml_data is not None:
            root = ET.fromstring(xml_data)
        else:
            tree = ET.parse(xml_file)
            root = tree.getroot()
    except EnvironmentError, e:
        log_error("Failed to open/read an XML file. file=%s, error=%s" % (xml_file, str(e)))
        raise
    except ExpatError, e:
        log_error("Invalid XML syntax in the provided XML. code=%s, line=%s, offset=%s, xml_data=%s, xml_file=%s" % (e.code, e.lineno, e.offset, xml_data, xml_file))
        raise FormatError("Invalid XML syntax in the provided XML.")

    if resource == "FileSystems":
        result = []
        for elm in root.findall('FileSystem'):
            result.append(FileSystemInfo(elm))
    elif resource == "FileSystem":
        result = FileSystemInfo(root)
    elif resource == "MountSetting":
        result = MountInfo(root)
    elif resource == "LargeFileTransfer":
        result = []
        for elm in root.findall('FileSystem'):
            result.append(LargeFileTransferInfo(elm))
    elif resource == "LargeFileTransferOne":
        result = LargeFileTransferInfo(root)
    elif resource == "CIFSShares":
        result = []
        for elm in root.findall('CIFSShare'):
            result.append(CIFSShareInfo(elm))
    elif resource == "CIFSShare":
        result = CIFSShareInfo(root)
    elif resource == "NFSShares":
        result = []
        for elm in root.findall('NFSShare'):
            result.append(NFSShareInfo(elm))
    elif resource == "NFSShare":
        result = NFSShareInfo(root)
    elif resource == "MigrationTasks":
        result = []
        for elm in root.findall('MigrationTask'):
            result.append(MigTaskInfo(elm))
    elif resource == "MigrationTask":
        result = MigTaskInfo(root)
    elif resource == "NetworkInterfaces":
        result = []
        # Get fixed IP addresses only
        for elm in root.findall('NetworkInterface'):
            fixedipelm = elm.find('FixedIP')
            ip = fixedipelm.find('target').text
            if ip is not None:
                result.append(ip)
    elif resource == "CacheResident":
        # Returns a boolean value. Not a list.
        try:
            e = root.find('CacheResident')
            result = Utility.str2bool(e.attrib['report'])
        except (AttributeError, ValueError, KeyError), e:
            log_error("Cache resident reporting setting is missing or wrong. error=%s" % str(e))
            raise FormatError("Wrong format of Cache resident reporting setting.")
    elif resource == "ProcessingNode":
        result = ProcessingNodeInfo(root)
    elif resource == "LUs":
        result = []
        for elm in root.findall('LU'):
            result.append(LUInfo(elm))
    elif resource == "VolumeGroups":
        result = []
        for elm in root.findall('VolumeGroup'):
            result.append(VolumeGroupInfo(elm))
    else:
        log_error("Unexpected resource was specified for _parse_xml(). resource=%s" % resource)
        raise ParameterError("Unexpected resource was specified for _parse_xml().")

    return result
